#! /usr/bin/env python
import os
import subprocess
import types
import tempfile
import shutil

import yaml
from distutils.dir_util import mkpath


class cd(object):
    def __init__(self, dir):
        self._dir = dir

    def __enter__(self):
        self._starting_dir = os.path.abspath(os.getcwd())
        if not os.path.isdir(self._dir):
            mkpath(self._dir)
        os.chdir(self._dir)
        return os.path.abspath(os.getcwd())

    def __exit__(self, type, value, traceback):
        os.chdir(self._starting_dir)


class cdtemp(object):
    def __init__(self, **kwds):
        self._kwds = kwds
        self._tmp_dir = None

    def __enter__(self):
        self._starting_dir = os.path.abspath(os.getcwd())
        self._tmp_dir = tempfile.mkdtemp(**self._kwds)
        os.chdir(self._tmp_dir)
        return os.path.abspath(self._tmp_dir)

    def __exit__(self, type, value, traceback):
        os.chdir(self._starting_dir)
        shutil.rmtree(self._tmp_dir)


class mktemp(object):
    def __init__(self, **kwds):
        self._kwds = kwds
        self._tmp_dir = None

    def __enter__(self):
        self._tmp_dir = tempfile.mkdtemp(**self._kwds)
        return os.path.abspath(self._tmp_dir)

    def __exit__(self, type, value, traceback):
        shutil.rmtree(self._tmp_dir)


def status(message):
    print ' '.join(['==>', message])


def check_output(*args, **kwds):
    kwds.setdefault('stdout', subprocess.PIPE)
    return subprocess.Popen(*args, **kwds).communicate()[0]


def system(*args, **kwds):
    verbose = kwds.pop('verbose', True)

    status(' '.join(args[0]))

    if verbose:
        call = subprocess.check_call
    else:
        call = check_output

    try:
        call(*args, **kwds)
    except subprocess.CalledProcessError:
        status('Error')
        raise


def which(prog, env=None):
    prog = os.environ.get(env or prog.upper(), prog)

    try:
        prog = check_output(['which', prog],
                            stderr=open('/dev/null', 'w')).strip()
    except subprocess.CalledProcessError:
        return None
    else:
        return prog


def pkg_config(name, opts):
    if isinstance(opts, types.StringTypes):
        opts = [opts]

    try:
        flags = check_output([which('pkg-config')] + opts + [name],
                             stderr=open('/dev/null', 'w')).strip()
    except subprocess.CalledProcessError:
        return None
    else:
        return flags



#! /usr/bin/env python

class Component(object):
    _input_var_names = set()
    _output_var_names = set()
    _var_units = dict()

    def __init__(self, grid, map_vars=None):
        map_vars = map_vars or {}
        self._grid = grid

        for (location, vars) in map_vars.items():
            for (dest, src) in vars.items():
                grid.add_field(location, dest,
                               grid.field_values(location, src))

    @property
    def input_var_names(self):
        return self._input_var_names

    @property
    def output_var_names(self):
        return self._output_var_names

    @property
    def name(self):
        return self._name

    @property
    def units(self):
        return self._var_units
    
    @property
    def var_units(self):
        return self._var_units
        
    @property
    def var_definitions(self):
        return self._var_defs
    
    @property
    def var_mapping(self):
        """var_mapping
        This is 'node', 'cell', 'active_link', etc.
        """
        return self._var_mapping

    @property
    def shape(self):
        return self.grid._shape

    @property
    def grid(self):
        return self._grid

    @property
    def coords(self):
        return (self.grid.node_x, self.grid.node_y)

    def imshow(self, name, **kwds):
        self._grid.imshow(name, **kwds)



#! /usr/env/python
"""
just a little script for testing the d8 flow routing class 
and drainage area calculation 
"""

from landlab import RasterModelGrid
from landlab.examples.flowRoutingD8 import RouteFlowD8
from landlab.examples.drainageArea import CalcDrainageArea
from numpy import *

def main():
    nr = 5
    nc = 6
    ncells = nr*nc
    dx=1
    #instantiate grid
    rg = RasterModelGrid(nr, nc, dx)

    elevations  = zeros( ncells )    
    #set-up interior elevations with random numbers
    #for i in range(0, ncells):
    #    if rg.is_interior(i):
    #        elevations[i]=random.random_sample()
    
    #set-up with prescribed elevations to test drainage area calcualtion
    helper = [7,13,19,10,16,22]
    for i in range(0, 6):
        #print 'helper[i]', helper[i]
        elevations[helper[i]]=2
    helper = [8,14,20,21,15,9]
    for i in range(0, 6):
        elevations[helper[i]]=3
    
    #tried making a pit, drainage area algorithm doesn't crash, but it doesn't
    #work perfectly either.    
    #elevations[15]=-50
    
    #printing elevations for debugging purposes
    print 'elevation vector' 
    print elevations

    #instantiate flow routing variable
    flow = RouteFlowD8(ncells)
    #calculate flow directions
    flow_directions = flow.calc_flowdirs(rg, elevations)    
    #printing flow directions for debugging purposes
    print 'flow direction vector'
    print flow_directions
    
    #instantiate drainage area variable
    da_calculator = CalcDrainageArea(ncells)
    #calculate drainage area
    drain_area = da_calculator.calc_DA(rg, flow_directions)
    #printing drainage area for debugging purposes
    print 'drainage area vector'
    print drain_area
    
    ##test of other code bit in raster model grid
    #[ms, ma] = rg.calculate_max_gradient_across_node(elevations,12)
    #print 'max slope', ms
    #print 'max angle', ma


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
This is an implementation of Vaughan Voller's experimental boundary method
reduced complexity flow router. Credit: Voller, Hobley, Paola.

Created on Fri Feb 20 09:32:27 2015

@author: danhobley (SiccarPoint), after volle001@umn.edu
"""

###in the diagonal case, even closed edges can produce "drag". Is this right?
#Could suppress by mirroring the diagonals

import numpy as np
from landlab import RasterModelGrid, ModelParameterDictionary, Component, FieldError
import inspect

class PotentialityFlowRouter(Component):
    """
    This class implements Voller, Hobley, and Paola's experimental matrix
    solutions for flow routing.
    Options are permitted to allow "abstract" routing (flow enforced downslope,
    but no particular assumptions are made about the governing equations), or
    routing according to the Chezy or Manning equations. This routine assumes
    that water is distributed evenly over the surface of the cell in deriving
    the depth, and does not assume channelization. You will need to back-
    calculate channel depths for yourself using known widths at each node
    if that is what you want.
    
    """
    _name = 'PotentialityFlowRouter'
    
    _input_var_names = set(['topographic_elevation',
                            'water__volume_flux_in',
                            ])
    
    _output_var_names = set(['water__volume_flux_magnitude',
                             'water__volume_flux_xcomponent',
                             'water__volume_flux_ycomponent',
                             'potentiality_field',
                             'water__depth',
                             'water__volume_flux',
                             ])
                             
    _var_units = {'topographic_elevation' : 'm',
                  'water__volume_flux_in' : 'm**3/s',
                  'water__volume_flux_magnitude' : 'm**3/s',
                  'water__volume_flux_xcomponent' : 'm**3/s',
                  'water__volume_flux_ycomponent' : 'm**3/s',
                  'potentiality_field' : 'm**3/s',
                  'water__depth': 'm',
                  'water__volume_flux' : 'm**3/s',
                  }
    
    _var_mapping = {'topographic_elevation' : 'node',
                  'water__volume_flux_in' : 'node',
                  'water__volume_flux_magnitude' : 'node',
                  'water__volume_flux_xcomponent' : 'node',
                  'water__volume_flux_ycomponent' : 'node',
                  'potentiality_field' : 'node',
                  'water__depth': 'node',
                  'water__volume_flux' : 'link',
                  }
    
    _var_defs = {'topographic_elevation' : 'Land surface topographic elevation',
                  'water__volume_flux_in' : 'External volume water input to each node (e.g., rainfall)',
                  'water__volume_flux_magnitude' : 'Magnitude of volumetric water flux through each node',
                  'water__volume_flux_xcomponent' : 'x component of resolved water flux through node',
                  'water__volume_flux_ycomponent' : 'y component of resolved water flux through node',
                  'potentiality_field' : 'Value of the hypothetical field "K", used to force water flux to flow downhill',
                  'water__depth': 'If Manning or Chezy specified, the depth of flow in the cell, calculated assuming flow occurs over the whole surface',
                  'water__volume_flux' : 'Water fluxes on links',
                  }
    
    _min_slope_thresh = 1.e-24 #if your flow isn't connecting up, this probably needs to be reduced

    
    def __init__(self, grid, params):
        self.initialize(grid, params)
    
    
    def initialize(self, grid, params):
        """
        Optional input parameters are:
        * "flow_equation" - options are "default" (default), "Manning", or "Chezy".
          If Equation is Manning or Chezy, you must also specify:
           - "Mannings_n" (if "Manning") : float
           - "Chezys_C" (if "Chezy") : float
        """
        assert RasterModelGrid in inspect.getmro(grid.__class__)
        assert grid.number_of_node_rows >= 3
        assert grid.number_of_node_columns >= 3
        
        self._grid = grid
        
        if type(params) == str:
            input_dict = ModelParameterDictionary(params)
        else:
            assert type(params) == dict
            input_dict = params
        
        #ingest the inputs
        try:
            self.equation = input_dict['flow_equation']
        except KeyError:
            self.equation = 'default'
        assert self.equation in ('default', 'Chezy', 'Manning')
        if self.equation == 'Chezy':
            self.chezy_C = float(input_dict["Chezys_C"])
        if self.equation == 'Manning':
            self.manning_n = float(input_dict["Mannings_n"])
        
        ncols = grid.number_of_node_columns
        nrows = grid.number_of_node_rows
        #create the blank node maps; assign the topo to an internally held 2D map with dummy edges:
        self.elev_raster = np.empty((nrows+2, ncols+2), dtype=float)
        self._aPP = np.zeros_like(self.elev_raster)
        self._aWW = np.zeros_like(self.elev_raster)
        self._aWP = np.zeros_like(self.elev_raster)
        self._aEE = np.zeros_like(self.elev_raster)
        self._aEP = np.zeros_like(self.elev_raster)
        self._aNN = np.zeros_like(self.elev_raster)
        self._aNP = np.zeros_like(self.elev_raster)
        self._aSS = np.zeros_like(self.elev_raster)
        self._aSP = np.zeros_like(self.elev_raster)
        self._uE = np.zeros_like(self.elev_raster)
        self._uW = np.zeros_like(self.elev_raster)
        self._uN = np.zeros_like(self.elev_raster)
        self._uS = np.zeros_like(self.elev_raster)
        self._uNE = np.zeros_like(self.elev_raster)
        self._uNW = np.zeros_like(self.elev_raster)
        self._uSE = np.zeros_like(self.elev_raster)
        self._uSW = np.zeros_like(self.elev_raster)
        self._K = np.zeros_like(self.elev_raster)
        
        #extras for diagonal routing:
        self._aNWNW = np.zeros_like(self.elev_raster)
        self._aNWP = np.zeros_like(self.elev_raster)
        self._aNENE = np.zeros_like(self.elev_raster)
        self._aNEP = np.zeros_like(self.elev_raster)
        self._aSESE = np.zeros_like(self.elev_raster)
        self._aSEP = np.zeros_like(self.elev_raster)
        self._aSWSW = np.zeros_like(self.elev_raster)
        self._aSWP = np.zeros_like(self.elev_raster)
        self._totalfluxout = np.empty((nrows, ncols), dtype=float)
        self._meanflux = np.zeros_like(self._totalfluxout)
        self._xdirfluxout = np.zeros_like(self._totalfluxout)
        self._ydirfluxout = np.zeros_like(self._totalfluxout)
        self._xdirfluxin = np.zeros_like(self._totalfluxout)
        self._ydirfluxin = np.zeros_like(self._totalfluxout)
        
        #setup slices for use in IDing the neighbors
        self._Es = (slice(1,-1),slice(2,ncols+2))
        self._NEs = (slice(2,nrows+2),slice(2,ncols+2))
        self._Ns = (slice(2,nrows+2),slice(1,-1))
        self._NWs = (slice(2,nrows+2),slice(0,-2))
        self._Ws = (slice(1,-1),slice(0,-2))
        self._SWs = (slice(0,-2),slice(0,-2))
        self._Ss = (slice(0,-2),slice(1,-1))
        self._SEs = (slice(0,-2),slice(2,ncols+2))
        self._core = (slice(1,-1),slice(1,-1))
        self._corecore = (slice(2,-2),slice(2,-2)) #the actual, LL-sense core (interior) nodes of the grid
        
        for out_field in self._output_var_names:
            if self._var_mapping[out_field]=='node':
                try:
                    self._grid.at_node[out_field]
                except FieldError:
                    self._grid.at_node[out_field] = np.empty(self._grid.number_of_nodes, dtype=float)
            elif self._var_mapping[out_field]=='link':
                try:
                    self._grid.at_link[out_field]
                except FieldError:
                    self._grid.at_link[out_field] = np.empty(self._grid.number_of_links, dtype=float)
        
        #make and store a 2d reference for node BCs
        self._BCs = 4*np.ones_like(self.elev_raster)
        self._BCs[self._core].flat = self._grid.get_node_status()
        BCR = self._BCs #for conciseness below
        #these are conditions for boundary->boundary contacts AND anything->closed contacts w/i the grid, both of which forbid flow
        self.boundaryboundaryN = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._Ns]>0), np.logical_or(BCR[self._Ns]==4, BCR[self._core]==4))
        self.boundaryboundaryS = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._Ss]>0), np.logical_or(BCR[self._Ss]==4, BCR[self._core]==4))
        self.boundaryboundaryE = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._Es]>0), np.logical_or(BCR[self._Es]==4, BCR[self._core]==4))
        self.boundaryboundaryW = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._Ws]>0), np.logical_or(BCR[self._Ws]==4, BCR[self._core]==4))
        self.boundaryboundaryNE = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._NEs]>0), np.logical_or(BCR[self._NEs]==4, BCR[self._core]==4))
        self.boundaryboundaryNW = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._NWs]>0), np.logical_or(BCR[self._NWs]==4, BCR[self._core]==4))
        self.boundaryboundarySE = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._SEs]>0), np.logical_or(BCR[self._SEs]==4, BCR[self._core]==4))
        self.boundaryboundarySW = np.logical_or(np.logical_and(BCR[self._core]>0, BCR[self._SWs]>0), np.logical_or(BCR[self._SWs]==4, BCR[self._core]==4))
        self.notboundaries = BCR[self._core] == 0
        
        self.equiv_circ_diam = 2.*np.sqrt(grid.dx*grid.dy/np.pi)
        #this is the equivalent seen CSWidth of a cell for a flow in a generic 360 direction
        
    
    def route_flow(self, route_on_diagonals=True, return_components=False):
        """
        """
        #create aliases for ease
        hR = self.elev_raster
        aPP = self._aPP
        aWW = self._aWW
        aWP = self._aWP
        aEE = self._aEE
        aEP = self._aEP
        aNN = self._aNN
        aNP = self._aNP
        aSS = self._aSS
        aSP = self._aSP
        totalfluxout = self._totalfluxout
        meanflux = self._meanflux
        uE = self._uE
        uW = self._uW
        uN = self._uN
        uS = self._uS
        K = self._K
        Es = self._Es
        NEs = self._NEs
        Ns = self._Ns
        NWs = self._NWs
        Ws = self._Ws
        SWs = self._SWs
        Ss = self._Ss
        SEs = self._SEs
        core = self._core
        one_over_dx = 1./self._grid.dx
        one_over_dy = 1./self._grid.dy
        one_over_diagonal = 1./np.sqrt(self._grid.dx**2+self._grid.dy**2)
        qwater_in = self._grid.at_node['water__volume_flux_in'].reshape((self._grid.number_of_node_rows, self._grid.number_of_node_columns))
        prev_K = K.copy()
        bbN = self.boundaryboundaryN
        bbS = self.boundaryboundaryS
        bbE = self.boundaryboundaryE
        bbW = self.boundaryboundaryW
        if return_components:
            flux_error = np.zeros_like(self._totalfluxout)
        
        if route_on_diagonals:
            #extras for diagonal routing:
            aNWNW = self._aNWNW
            aNWP = self._aNWP
            aNENE = self._aNENE
            aNEP = self._aNEP
            aSESE = self._aSESE
            aSEP = self._aSEP
            aSWSW = self._aSWSW
            aSWP = self._aSWP
            
            bbNE = self.boundaryboundaryNE
            bbNW = self.boundaryboundaryNW
            bbSE = self.boundaryboundarySE
            bbSW = self.boundaryboundarySW            
            
        #paste in the elevs
        hR[1:-1,1:-1].flat = self._grid.at_node['topographic_elevation']
        
        #update the dummy edges of our variables - these all act as closed nodes (the inner, true boundaries are handled elsewhere... or they should be closed anyway!):
        #note this isn't sufficient of we have diagonals turned on, as flow can still occur on them
        hR[0,1:-1] = hR[1,1:-1]
        hR[-1,1:-1] = hR[-2,1:-1]
        hR[1:-1,0] = hR[1:-1,1]
        hR[1:-1,-1] = hR[1:-1,-2]
        hR[(0,-1,0,-1),(0,-1,-1,0)] = hR[(1,-2,1,-2),(1,-2,-2,1)]
        
        aNN[core] = (-hR[core]+hR[Ns]).clip(0.)
        aNP[core] =  (hR[core]-hR[Ns]).clip(0.)
        aSS[core] = (-hR[core]+hR[Ss]).clip(0.)
        aSP[core] =  (hR[core]-hR[Ss]).clip(0.)
        aEE[core] = (-hR[core]+hR[Es]).clip(0.)
        aEP[core] =  (hR[core]-hR[Es]).clip(0.)
        aWW[core] = (-hR[core]+hR[Ws]).clip(0.)
        aWP[core] =  (hR[core]-hR[Ws]).clip(0.)
        aNN[core] *= one_over_dy
        aNP[core] *= one_over_dy
        aSS[core] *= one_over_dy
        aSP[core] *= one_over_dy
        aEE[core] *= one_over_dx
        aEP[core] *= one_over_dx
        aWW[core] *= one_over_dx
        aWP[core] *= one_over_dx
        
        #disable lateral flow betw boundary nodes:
        aEE[core][bbE] = 0.
        aEP[core][bbE] = 0.
        aWW[core][bbW] = 0.
        aWP[core][bbW] = 0.
        aNN[core][bbN] = 0.
        aNP[core][bbN] = 0.
        aSS[core][bbS] = 0.
        aSP[core][bbS] = 0.
        #diag correction happens below
        
        if self.equation != 'default':
            for grad in (aEE,aEP,aWW,aWP,aNN,aNP,aSS,aSP):
                np.sqrt(grad[core], out=grad[core]) #...because both Manning and Chezy actually follow sqrt slope, not slope
        
        if route_on_diagonals:
            #adding diagonals
            aNENE[core] = (-hR[core]+hR[NEs]).clip(0.)
            aNEP[core] =   (hR[core]-hR[NEs]).clip(0.)
            aSESE[core] = (-hR[core]+hR[SEs]).clip(0.)
            aSEP[core] =   (hR[core]-hR[SEs]).clip(0.)
            aSWSW[core] = (-hR[core]+hR[SWs]).clip(0.)
            aSWP[core] =   (hR[core]-hR[SWs]).clip(0.)
            aNWNW[core] = (-hR[core]+hR[NWs]).clip(0.)
            aNWP[core] =   (hR[core]-hR[NWs]).clip(0.)
            aNENE[core] *= one_over_diagonal
            aNEP[core] *= one_over_diagonal
            aSESE[core] *= one_over_diagonal
            aSEP[core] *= one_over_diagonal
            aSWSW[core] *= one_over_diagonal
            aSWP[core] *= one_over_diagonal
            aNWNW[core] *= one_over_diagonal
            aNWP[core] *= one_over_diagonal
            
            #disable lateral flow betw boundaries...
            aNENE[core][bbNE] = 0.
            aNEP[core][bbNE] = 0.
            aNWNW[core][bbNW] = 0.
            aNWP[core][bbNW] = 0.
            aSESE[core][bbSE] = 0.
            aSEP[core][bbSE] = 0.
            aSWSW[core][bbSW] = 0.
            aSWP[core][bbSW] = 0.
            
            if self.equation != 'default':
                for grad in (aNENE,aNEP,aNWNW,aNWP,aSESE,aSEP,aSWSW,aSWP):
                    np.sqrt(grad[core], out=grad[core]) #...because both Manning and Chezy actually follow sqrt slope, not slope
                
        if not route_on_diagonals:
            aPP[core] = aWP[core]+aEP[core]+aSP[core]+aNP[core]+self._min_slope_thresh
        else:
            aPP[core] = (aWP[core]+aEP[core]+aSP[core]+aNP[core]
                        +aNEP[core]+aSEP[core]+aSWP[core]+aNWP[core])+self._min_slope_thresh
                        
        mismatch = 10000.
        self.loops_needed = 0

        #this explicit solution could easily be replaced with a better solver        
        while mismatch>1.e-3:
            if not route_on_diagonals:
                K[core] = (aWW[core]*K[Ws]+aEE[core]*K[Es]+aSS[core]*K[Ss]+aNN[core]*K[Ns]
                                    +qwater_in)/aPP[core]
            else:
                K[core] = (aWW[core]*K[Ws]+aEE[core]*K[Es]+aSS[core]*K[Ss]+aNN[core]*K[Ns]
                          +aNENE[core]*K[NEs]+aSESE[core]*K[SEs]
                          +aSWSW[core]*K[SWs]+aNWNW[core]*K[NWs]
                          +qwater_in)/aPP[core]

            mismatch = np.sum(np.square(K[core]-prev_K[core]))
            self.loops_needed += 1
            prev_K = K.copy()
            #print mismatch
            
            for BC in (K,):
                BC[0,1:-1] = BC[1,1:-1]
                BC[-1,1:-1] = BC[-2,1:-1]
                BC[1:-1,0] = BC[1:-1,1]
                BC[1:-1,-1] = BC[1:-1,-2]
                BC[(0,-1,0,-1),(0,-1,-1,0)] = BC[(1,-2,1,-2),(1,-2,-2,1)]
        
        if route_on_diagonals:
            outdirs = (aNP,aSP,aEP,aWP,aNWP,aNEP,aSWP,aSEP)
            indirs = ((aNN,K[Ns]),(aSS,K[Ss]),(aEE,K[Es]),(aWW,K[Ws]),(aNWNW,K[NWs]),(aNENE,K[NEs]),(aSWSW,K[SWs]),(aSESE,K[SEs]))
        else:
            outdirs = (aNP,aSP,aEP,aWP)
            indirs = ((aNN,K[Ns]),(aSS,K[Ss]),(aEE,K[Es]),(aWW,K[Ws]))
        
        totalfluxout.fill(0.)
        for array in outdirs:
            totalfluxout += array[core]
        totalfluxout *= K[core]        
        no_outs = np.equal(totalfluxout,0.)
        #this WON'T work if there is no out flux, i.e., a BC, so - 
        if np.any(no_outs):
            for (inarray, inK) in indirs:
                totalfluxout[no_outs] += inarray[core][no_outs]*inK[no_outs]
        yes_outs = np.logical_not(no_outs)
        meanflux[:] = totalfluxout
        meanflux[yes_outs] -= 0.5*qwater_in[yes_outs]
        #so note the BC nodes get the value of the fluxes they RECEIVE, without any addition of flux
        #& also, nodes at the TOP of the network, with no explicit ins, get averaged values, not just their out values
            
        
        if return_components:
            #this takes a nontrivial number of additional calculations, so is optional
            #we aim to return the MEAN flow direction.
            if route_on_diagonals:
                prefactor_to_x = np.cos(np.arctan(self._grid.dy/self._grid.dx))
                prefactor_to_y = np.sin(np.arctan(self._grid.dy/self._grid.dx))
                xdir_mod_out = (0.,0.,1.,-1.,-prefactor_to_x,prefactor_to_x,-prefactor_to_x,prefactor_to_x)
                ydir_mod_out = (1.,-1.,0.,0.,prefactor_to_y,prefactor_to_y,-prefactor_to_y,-prefactor_to_y)
                xdir_mod_in = (0.,0.,-1.,1.,prefactor_to_x,-prefactor_to_x,prefactor_to_x,-prefactor_to_x)
                ydir_mod_in = (-1.,1.,0.,0.,-prefactor_to_y,-prefactor_to_y,prefactor_to_y,prefactor_to_y)
            else:
                xdir_mod_out = (0.,0.,1.,-1.)
                ydir_mod_out = (1.,-1.,0.,0.)
                xdir_mod_in = (0.,0.,-1.,1.)
                ydir_mod_in = (-1.,1.,0.,0.)
            #out is easier, so do it first:
            self._xdirfluxout.fill(0.)
            self._ydirfluxout.fill(0.)
            for (array,mod) in zip(outdirs,xdir_mod_out):
                self._xdirfluxout += mod*array[core]
            for (array,mod) in zip(outdirs,ydir_mod_out):
                self._ydirfluxout += mod*array[core]
            #now in
            self._xdirfluxin.fill(0.)
            self._ydirfluxin.fill(0.)
            for ((array,Kin),mod) in zip(indirs,xdir_mod_in):
                self._xdirfluxin += mod*array[core]*Kin
            for ((array,Kin),mod) in zip(indirs,ydir_mod_in):
                self._ydirfluxin += mod*array[core]*Kin
            #modify for doing the means:
            self._xdirfluxin[yes_outs] *= 0.5
            self._ydirfluxin[yes_outs] *= 0.5
            #flux outs all get adjusted:
            self._xdirfluxout *= 0.5
            self._ydirfluxout *= 0.5
            #NB: handling of degenerate cases where there is no angle defined (though rare) could be problematic.
            # => assign an arbitary zero angle in these cases
            #so...
            mean_x = self._xdirfluxin + self._xdirfluxout
            mean_y = self._ydirfluxin + self._ydirfluxout
            apparent_flux = np.sqrt(mean_x*mean_x + mean_y*mean_y)
            degenerate_fluxes = np.equal(apparent_flux,0.)
            if np.any(degenerate_fluxes):
                #THIS PART HAS NOT BEEN TESTED
                defined_fluxes = np.logical_not(degenerate_fluxes)
                flux_error[defined_fluxes] = meanflux[defined_fluxes]/apparent_flux[defined_fluxes]
                #zeros in the degenerate slots
                self._grid.at_node['water__volume_flux_xcomponent'][:] = (mean_x*flux_error).flat
                #now modify so we get the right answer for total flux, pointing arbitrarily N
                self._grid.at_node['water__volume_flux_ycomponent'][defined_fluxes] = (mean_y[defined_fluxes]*flux_error[defined_fluxes]).flat
                self._grid.at_node['water__volume_flux_ycomponent'][degenerate_fluxes] = meanflux[degenerate_fluxes].flat
            else:
                flux_error[:] = meanflux/apparent_flux
                self._grid.at_node['water__volume_flux_ycomponent'][:] = (mean_y*flux_error).flat
                self._grid.at_node['water__volume_flux_xcomponent'][:] = (mean_x*flux_error).flat
            
        #save the output
        self._grid.at_link['water__volume_flux'][self._grid.node_links()[0]] = uS[core].flat #[S,W,N,E], (4,nnodes)
        self._grid.at_link['water__volume_flux'][self._grid.node_links()[1]] = uW[core].flat
        self._grid.at_link['water__volume_flux'][self._grid.node_links()[2]] = uN[core].flat
        self._grid.at_link['water__volume_flux'][self._grid.node_links()[3]] = uE[core].flat
        self._grid.at_node['potentiality_field'][:] = K[core].flat
        self._grid.at_node['water__volume_flux_magnitude'][:] = meanflux.flat
        #the x,y components are created above, in the if statement
        
        #now process uval and vval to give the depths, if Chezy or Manning:
        if self.equation == 'Chezy':
            #Chezy: Q = C*Area*sqrt(depth*slope)
            self._grid.at_node['water__depth'][:] = (self._grid.at_node['potentiality_field']/self.chezy_C/self.equiv_circ_diam)**(2./3.)
        elif self.equation == 'Manning':
            #Manning: Q = w/n*depth**(5/3)
            self._grid.at_node['water__depth'][:] = (self._grid.at_node['potentiality_field']*self.manning_n/self.equiv_circ_diam)**0.6
        else:
            pass


import numpy as np
from landlab import ModelParameterDictionary
from time import sleep
from scipy import weave
from scipy.weave.build_tools import CompileError

from landlab.core.model_parameter_dictionary import MissingKeyError
from landlab.field.scalar_data_fields import FieldError

class TransportLimitedEroder(object):
    """
    This component implements transport limited erosion for a landscape in which
    flow directions are fully convergent. i.e., all nodes in the landscape have
    a single, uniquely defined downstream node.
    
    The module can in principle take multiple transport laws, but at the moment
    only Meyer-Peter Muller (MPM) is implemented.
    
    There is as yet no explicit stabilization check on the timestep. If your
    run destabilizes, try reducing dt.
    See, e.g., ./examples/simple_sp_driver.py
    
    Assumes grid does not deform or change during run.
    
    Note it is vital to ensure all your units match. t is assumed to be in 
    years. Any length value is assumed to be in meters (including both dx
    and the uplift rate...!)
    
    DEJH Sept 14.
    This component *should* run on any grid, but untested.
    """
    
    def __init__(self, grid, params):
        self.initialize(grid, params)
            
#This draws attention to a potential problem. It will be easy to have modules update z, but because noone "owns" the data, to forget to also update dz/dx...
#How about a built in grid utility that updates "derived" data (i.e., using only grid fns, e.g., slope, curvature) at the end of any given tstep loop?
#Or an explicit flagging system for all variables in the modelfield indicating if they have been updated this timestep. (Currently implemented)
#Or wipe the existance of any derived grid data at the end of a timestep entirely, so modules find they don't have it next timestep.
        
    def initialize(self, grid, params_file):
        '''
        params_file is the name of the text file containing the parameters 
        needed for this stream power component.
        
        ***Parameters for input file***
        OBLIGATORY:
            * Qc -> String. Controls how to set the carrying capacity.
                Either 'MPM', or a string giving the name of the model field
                where capacity values are stored on nodes.
                At the moment, only 'MPM' is permitted as a way to set the 
                capacity automatically, but expansion would be trivial.
                If 'from_array', the module will attempt to set the capacity
                Note capacities must be specified as volume flux.
            * 
            
            ...Then, assuming you set Qc=='MPM':
            * b_sp, c_sp -> Floats. These are the powers on discharge and 
                drainage area in the equations used to control channel width and 
                basin hydrology, respectively:
                        W = k_w * Q**b_sp
                        Q = k_Q * A**c_sp
                These parameters are used to constrain flow depth, and may be
                omitted if use_W or use_Q are set.
            *k_Q, k_w, mannings_n -> floats. These are the prefactors on the 
                basin hydrology and channel width-discharge relations, and n
                from the Manning's equation, respectively. These are 
                needed to allow calculation of shear stresses and hence carrying
                capacities from the local slope and drainage area alone.
                The equation for depth used to derive shear stress and hence
                carrying capacity contains a prefactor:
                    mannings_n*(k_Q**(1-b)/K_w)**0.6 
                (so shear = fluid_density*g*depth_equation_prefactor*A**(0.6*c*(1-b)*S**0.7 !)
                Don't know what to set these values to? k_w=0.002, k_Q=1.e-9,
                mannings_n=0.03 give vaguely plausible numbers (e.g., for a
                drainage area ~400km2, like Boulder Creek at Boulder,
                => depth~2.5m, width~35m, shear stress ~O(1000Pa)).
            *Dchar -> float.  The characteristic grain diameter in meters
                (==D50 in most cases) used to calculate Shields numbers
                in the channel. If you want to define Dchar values at each node,
                don't set, and use the Dchar_if_used argument in erode() 
                instead.
            
        OPTIONS:
            *rock_density -> in kg/m3 (defaults to 2700)
            *sediment_density -> in kg/m3 (defaults to 2700)
            *fluid_density -> in most cases water density, in kg/m3 (defaults to 1000)
            *g -> acceleration due to gravity, in m/s**2 (defaults to 9.81)
            
            *threshold_shields -> +ve float; the threshold taustar_crit. 
                Defaults to 0.047, or if 'slope_sensitive_threshold' is set True,
                becomes a weak function of local slope following Lamb et al 
                (2008):
                    threshold_shields=0.15*S**0.25
            *slope_sensitive_threshold -> bool, defaults to 'False'.
                If true, threshold_shields is set according to the Lamb
                equation. An exception will be raised if threshold_shields is 
                also set.
            *dt -> +ve float. If set, this is the fixed timestep for this
                component. Can be overridden easily as a parameter in erode(). 
                If not set (default), this parameter MUST be set in erode().
            *use_W -> Bool; if True, component will look for node-centered data
                describing channel width in grid.at_node['channel_width'], and 
                use it to implement incision ~ stream power per unit width. 
                Defaults to False.
            *use_Q -> Bool. Overrides the basin hydrology relation, using an
                local water discharge value assumed already calculated and
                stored in grid.at_node['discharge'].
            *C_MPM -> float. Defaults to 1. Allows tuning of the MPM prefactor,
                which is calculated as
                    Qc = 8.*C_MPM*(taustar - taustarcrit)**1.5
                In almost all cases, tuning depth_equation_prefactor' is 
                preferred to tuning this parameter.
            
        '''
        self.grid = grid
        self.link_S_with_trailing_blank = np.zeros(grid.number_of_links+1) #needs to be filled with values in execution
        self.count_active_links = np.zeros_like(self.link_S_with_trailing_blank, dtype=int)
        self.count_active_links[:-1] = 1
        inputs = ModelParameterDictionary(params_file)
        try:
            self.g = inputs.read_float('g')
        except MissingKeyError:
            self.g = 9.81
        try:
            self.rock_density = inputs.read_float('rock_density')
        except MissingKeyError:
            self.rock_density = 2700.
        try:
            self.sed_density = inputs.read_float('sediment_density')
        except MissingKeyError:
            self.sed_density = 2700.
        try:
            self.fluid_density = inputs.read_float('fluid_density')
        except MissingKeyError:
            self.fluid_density = 1000.
        self.relative_weight = (self.sed_density-self.fluid_density)/self.fluid_density*self.g #to accelerate MPM calcs
        self.excess_SG = (self.sed_density-self.fluid_density)/self.fluid_density
        self.rho_g = self.fluid_density*self.g     
        
        try:
            self.Qc = inputs.read_string('Qc')
        except MissingKeyError:
            raise MissingKeyError("Qc must be 'MPM' or a grid field name!")
        else:
            if self.Qc=='MPM':
                self.calc_cap_flag = True
            else:
                self.calc_cap_flag = False
                
        try:
            self.lamb_flag = inputs.read_bool('slope_sensitive_threshold')
        except:
            self.lamb_flag = False
        try:
            self.shields_crit = inputs.read_float('threshold_shields')
            self.set_threshold = True #flag for sed_flux_dep_incision to see if the threshold was manually set.
            print "Found a threshold to use: ", self.shields_crit
            assert self.lamb_flag == False
        except MissingKeyError:
            if not self.lamb_flag:
                self.shields_crit = 0.047
            self.set_threshold = False
        try:
            self.tstep = inputs.read_float('dt')
        except MissingKeyError:
            pass
        try:
            self.use_W = inputs.read_bool('use_W')
        except MissingKeyError:
            self.use_W = False
        try:
            self.use_Q = inputs.read_bool('use_Q')
        except MissingKeyError:
            self.use_Q = False
            
        try:
            self._b = inputs.read_float('b_sp')
        except MissingKeyError:
            if self.use_W:
                self._b = 0.
            else:
                if self.calc_cap_flag:
                    raise NameError('b was not set')
        try:
            self._c = inputs.read_float('c_sp')
        except MissingKeyError:
            if self.use_Q:
                self._c = 1.
            else:
                if self.calc_cap_flag:
                    raise NameError('c was not set')
        try:
            self.Dchar_in = inputs.read_float('Dchar')
        except MissingKeyError:
            pass
            
        #assume Manning's equation to set the power on A for shear stress:
        self.shear_area_power = 0.6*self._c*(1.-self._b)
        
        try:
            self.k_Q = inputs.read_float('k_Q')
        except MissingKeyError:
            self.depth_prefactor = self.rho_g*inputs.read_float('depth_equation_prefactor')
        else:
            self.k_w = inputs.read_float('k_w')
            mannings_n = inputs.read_float('mannings_n')
            if mannings_n<0. or mannings_n>0.2:
                print "***STOP. LOOK. THINK. You appear to have set Manning's n outside it's typical range. Did you mean it? Proceeding...***"
                sleep(2)
            self.depth_prefactor = self.rho_g*mannings_n*(self.k_Q**(1.-self._b)/self.k_w)**0.6
            ##Note the depth_prefactor we store already holds rho*g   

        try:
            self.C_MPM = inputs.read_float('C_MPM')
        except MissingKeyError:
            self.C_MPM = 1.
        try:
            self.shields_prefactor = 1./((self.sed_density-self.fluid_density)*self.g*self.Dchar_in)
            self.MPM_prefactor = 8.*self.C_MPM*np.sqrt(self.relative_weight*self.Dchar_in*self.Dchar_in*self.Dchar_in)
            self.MPM_prefactor_alt = 4.*self.g**(-2./3.)/self.excess_SG/self.fluid_density/self.sed_density
        except AttributeError:
            #have to set these manually as needed
            self.shields_prefactor_noD = 1./((self.sed_sensity-self.fluid_density)*self.g)

        self.cell_areas = np.empty(grid.number_of_nodes)
        self.cell_areas.fill(np.mean(grid.cell_areas))
        self.cell_areas[grid.cell_node] = grid.cell_areas
        
        
    def sed_capacity_equation(self, grid, shields_stress, slopes_at_nodes=None, areas_at_node=None):
        """
        slopes_at_nodes must be provided as nnodes array if slope_sensitive_threshold.
        This is a volume flux.
        At the moment, this only supports MPM for sed flux capacity calculation,
        but is set up to allow easy additions later if needed.
        """
        if self.lamb_flag:
            thresh = 0.15*slopes_at_nodes**0.25
        else:
            thresh = self.shields_crit
        if self.Qc == 'MPM':
            print "in capacity:"
            print 'prefactor ', self.MPM_prefactor
            print 'max Shields ', np.amax(shields_stress[self.grid.core_nodes])
            print 'thresh ', np.amax(thresh)
            try:
                capacity = self.MPM_prefactor*((shields_stress-thresh).clip(0.))**1.5
            except AttributeError:
                capacity = 8.*self.C_MPM*np.sqrt(self.relative_weight*self.Dchar*self.Dchar*self.Dchar)*((shields_stress-thresh).clip(0.))**1.5
        elif type(self.Qc)==str:
            try:
                capacity = grid.at_node[self.Qc]
            except:
                raise TypeError('sed_flux_dep_incision does not know how to set sed transport capacity from the provided input...')
        elif type(self.Qc)==np.ndarray:
            capacity=self.Qc
        else:
            raise TypeError('sed_flux_dep_incision does not know how to set sed transport capacity from the provided inputs...')
        capacity *= self.k_w*areas_at_node**self._b*31557600.
        return capacity #returned capacity is volume flux/yr
    
    def sed_capacity_equation_alt_method(self, grid, shear_stress, slopes_at_nodes=None, areas_at_node=None):
        """
        slopes_at_nodes must be provided as nnodes array if slope_sensitive_threshold.
        This is a volume flux.
        At the moment, this only supports MPM for sed flux capacity calculation,
        but is set up to allow easy additions later if needed.
        """
        if self.lamb_flag:
            thresh = 0.15*slopes_at_nodes**0.25
        else:
            thresh = self.shields_crit
        if self.Qc == 'MPM':
            capacity = self.MPM_prefactor_alt*((shear_stress-self.excess_SG*self.Dchar*self.rho_g*thresh).clip(0.))
        elif type(self.Qc)==str:
            try:
                capacity = grid.at_node[self.Qc]
            except:
                raise TypeError('sed_flux_dep_incision does not know how to set sed transport capacity from the provided input...')
        elif type(self.Qc)==np.ndarray:
            capacity=self.Qc
        else:
            raise TypeError('sed_flux_dep_incision does not know how to set sed transport capacity from the provided inputs...')
        capacity *= self.k_w*areas_at_node**self._b*31557600.
        return capacity #returned capacity is volume flux/yr
    
    
    def weave_iter_sed_trp_balance(self, grid, capacity, r, s, dt):
        '''
        Needs to be passed at node values for 'flow_receiver' and 
        'upstream_ID_order' (i.e., r and s). These are most likely the output
        values from route_flow_dn.
        Returns dz/dt, the rate of elevation change.
        '''
        code =  """
                int donor;
                int rcvr;
                for (int i = num_pts-1; i > -1; i--) {
                    donor = s[i];
                    rcvr = r[donor];
                    dz[donor] = (sed_into_cell[donor] - capacity[donor]*dt)/cell_areas[donor];
                    if (donor != rcvr) {
                        sed_into_cell[rcvr] += capacity[donor];
                    }
                }
                """
        num_pts = len(s)
        cell_areas = self.cell_areas #nnodes long
        sed_into_cell = np.zeros(capacity.size, dtype=np.float64) #this is, more accurately, a sed *flux* in
        dz = np.zeros_like(sed_into_cell)
        #print type(cell_areas), type(s), type(r), type(capacity), type(sed_into_cell)
        try:
            raise CompileError #force the python loop for debug
            weave.inline(code, ['num_pts', 'cell_areas', 's', 'r', 'capacity', 'sed_into_cell', 'dz', 'dt'])
        except CompileError:
            for donor in s[::-1]:
                rcvr = r[donor]
                #by the time we get to donor, no more sed will be added; it's "done":
    ########where should the next line be? Can an internally drained node erode down/deposit?
                dz[donor] = (sed_into_cell[donor]-capacity[rcvr]*dt)/cell_areas[donor] #if capacity is greater than sed_in, it will erode (-ve value here)
                #we look downstream to ensure continuity with the BCs at grid margins
                #print "dz: ", dz[donor]
                #capacity's worth of sed leaves donor & goes to rcvr
                if donor != rcvr:
                    #print "saved", donor, rcvr
                    #print "dz: ", dz[donor]
                    sed_into_cell[rcvr] += capacity[donor]
        #print dz
        return dz

        
    def erode(self, grid, dt, node_drainage_areas='drainage_area', 
                slopes_at_nodes=None, link_slopes=None, link_node_mapping='links_to_flow_receiver', 
                receiver='flow_receiver', upstream_order='upstream_ID_order', 
                slopes_from_elevs=None, W_if_used=None, Q_if_used=None,
                Dchar_if_used=None, io=None):
        
        """
        This method calculates the fluvial sediment transport capacity at all
        nodes, then erodes or deposits sediment as required.
        
        *grid* & *dt* are the grid object and timestep (float) respectively.
        
        *Node_drainage_areas* tells the component where to look for the drainage
        area values. Change to another string to override which grid field the
        component looks at, or pass a nnodes-long array of drainage areas values
        directly instead.
        
        If you already have slopes defined at nodes on the grid, pass them to
        the component with *slopes_at_nodes*. The same syntax is expected: 
        string gives a name in the grid fields, an array gives values direct.
        
        Alternatively, set *link_slopes* (and *link_node_mapping*) if this data
        is only available at links. 'planet_surface__derivative_of_elevation'
        is the default field name for link slopes. Override this name by
        setting the variable as the appropriate string, or override use of 
        grid fields altogether by passing an array. *link_node_mapping* controls
        how the component maps these link values onto the arrays. We assume 
        there is always a 1:1 mapping (pass the values already projected onto 
        the nodes using slopes_at_nodes if not). Other components, e.g.,
        flow_routing.route_flow_dn, may provide the necessary outputs to make
        the mapping easier: e.g., just pass 'links_to_flow_reciever' from that
        module (the default name). If the component cannot find an existing
        mapping through this parameter, it will derive one on the fly, at
        considerable cost of speed (see on-screen reports).
        
        *slopes_from_elevs* allows the module to create gradients internally
        from elevations rather than have them provided. Set to True to force 
        the component to look for the data in grid.at_node['topographic_elevation'];
        set to 'name_of_field' to override this name, or pass an nnode-array
        to use those values as elevations instead. Using this option is 
        considerably slower than any of the alternatives, as it also has to 
        calculate the link_node_mapping from stratch each time.
        
        In both these cases, at present the mapping is to use the maximum
        slope of --any-- link attached to the node as the representative node
        slope. This is primarily for speed, but may be a good idea to modify
        later.
        
        *W_if_used* and *Q_if_used* must be provided if you set use_W and use_Q
        respectively in the component initialization. They can be either field
        names or nnodes arrays as in the other cases.
        
        *Dchar_if_used* must be set as a grid field string or nnoodes-long array
        if 'Dchar' as a float was not provided in the input file. (If it was,
        this will be overridden).
        
        RETURNS XXXXXX
        """
        
        try:
            r_in = grid.at_node[receiver]
        except FieldError:
            r_in = receiver
            assert r_in.size == grid.number_of_nodes
        try:
            s_in = grid.at_node[upstream_order]
        except FieldError:
            s_in = upstream_order
            assert s_in.size == grid.number_of_nodes
        try:
            self.Dchar=self.Dchar_in
        except AttributeError:
            try:
                self.Dchar=grid.at_node[Dchar_if_used]
            except FieldError:
                assert type(Dchar_if_used)==np.ndarray
                self.Dchar=Dchar_if_used
            
        #Perform check on whether we use grid or direct fed data:
        if slopes_at_nodes==None:
            if slopes_from_elevs:
                if slopes_from_elevs == True:
                    node_z = grid.at_node['topographic_elevation']
                elif type(slopes_from_elevs) == str:
                    node_z = grid.at_node[slopes_from_elevs]
                else:
                    node_z = slopes_from_elevs
                S_links = (node_z[grid.node_index_at_link_head]-node_z[grid.node_index_at_link_tail])/grid.link_length
            else:
                if link_slopes:
                    if type(link_slopes)==str:
                        S_links = grid.at_link[link_slopes]
                    else:
                        S_links = link_slopes
                else:
                    S_links = grid.at_link['planet_surface__derivative_of_elevation']
                    
            #put the slopes onto the nodes
            try:
                self.slopes = S_links[grid.at_node[link_node_mapping]]
            except FieldError:
                try:
                    self.slopes = S_links[link_node_mapping]
                except IndexError:
                    #need to do the mapping on the fly.
                    #we're going to use the max slope (i.e., + or -) of *all* adjacent nodes.
                    #This isn't ideal. It should probably just be the outs...
                    #i.e., np.amax(self.link_S_with_trailing_blank[grid.node_outlinks] AND -self.link_S_with_trailing_blank[grid.node_inlinks])
                    self.link_S_with_trailing_blank[:-1] = S_links
                    self.slopes = np.amax(np.fabs(self.link_S_with_trailing_blank[grid.node_links]),axis=0)
        else:
            try:
                self.slopes = grid.at_node[slopes_at_nodes]
            except FieldError:
                self.slopes = slopes_at_nodes
        
        #print 'Max slope ', np.amax(self.slopes[self.grid.core_nodes])
        #print 'Mean slope ', np.mean(self.slopes[self.grid.core_nodes])
        if type(node_drainage_areas)==str:
            node_A = grid.at_node[node_drainage_areas]
        else:
            node_A = node_drainage_areas

        #print 'Area ', np.amax(node_A)
        #calc Shields
######need to add ability to do this with W or Q scaling specified
        shear_stress = self.depth_prefactor*node_A**self.shear_area_power*self.slopes**0.7
        try:
            shields_stress = self.shields_prefactor*shear_stress
        except AttributeError:
            shields_stress = self.shields_prefactor_noD/self.Dchar*shear_stress
        #print 'max_shear ', np.amax(shear_stress[self.grid.core_nodes])
        #capacity = self.sed_capacity_equation(grid, shields_stress, self.slopes, node_A)
        capacity = self.sed_capacity_equation_alt_method(grid, shear_stress, self.slopes, node_A)
        #print capacity
        #need to strip closed nodes from s:
        #...but in fact we use the mask on capacity, not a BC check, as some open nodes can still have no slopes defined (so are functionally closed) - i.e., the corners
        try:
            s_in = np.delete(s_in, np.where(np.in1d(s_in,np.where(capacity.mask),assume_unique=True))) #yuck! #this isn't very memory efficient...
        except AttributeError: #provided S array might not be masked...
            s_in = np.delete(s_in, np.where(self.grid.is_boundary(s_in, boundary_flag=4)))
        #print "r: ", r_in
        #print "s: ", s_in
        dz = self.weave_iter_sed_trp_balance(grid, capacity, r_in, s_in, dt)
        #print 'max inc rate ', np.amax(np.fabs(dz[self.grid.core_nodes]))
        active_nodes = grid.get_active_cell_node_ids()
        if io:
            try:
                io[active_nodes] += dz[active_nodes]
            except TypeError:
                if type(io)==str:
                    elev_name = io
            else:
                return grid, io, capacity
            
        else:
            elev_name = 'topographic_elevation'

        grid.at_node[elev_name][active_nodes] += dz[active_nodes]
        grid.at_node['sediment_flux_capacity'] = capacity
        
        return grid, grid.at_node[elev_name], capacity
        



import numpy as np


from ...utils.jaggedarray import JaggedArray
from .status import CORE_NODE, CLOSED_BOUNDARY
from .links import _split_link_ends as _split_face_ends


def link_is_active(status_at_link_ends):
    """link_is_active((status0, status1))
    Check if a link is active.

    Links are *inactive* if they connect two boundary nodes or touch a
    closed boundary. Otherwise, the link is *active*.

    Parameters
    ----------
    status0, status1 : sequence of array-like
        Status at link start and end

    Returns
    -------
    ndarray, boolean :
        Boolean array that indicates if a link is active.
    """
    (status_at_link_start,
     status_at_link_end) = _split_link_ends(status_at_link_ends)

    return (((status_at_link_start == CORE_NODE) &
             ~ (status_at_link_end == CLOSED_BOUNDARY)) |
            ((status_at_link_end == CORE_NODE) &
             ~ (status_at_link_start == CLOSED_BOUNDARY)))


def find_active_links(node_status, node_at_link_ends):
    """find_active_links(node_status, (node0, node1))
    IDs of active links.

    Parameters
    ----------
    node_status : ndarray
        Status of nodes.
    node0, node1 : sequence of array-like
        Node ID at link start and end.

    Returns
    -------
    ndarray :
        Links IDs of active links.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import find_active_links
    >>> links = [(0, 2), (1, 3), (0, 1), (1, 2), (0, 3)]
    >>> status = np.array([0, 0, 0, 0])
    >>> find_active_links(status, links)
    array([0, 1, 2, 3, 4])
    """
    node_at_link_start, node_at_link_end = _split_link_ends(node_at_link_ends)

    if len(node_at_link_end) != len(node_at_link_start):
        raise ValueError('Link arrays must be the same length')

    status_at_link_ends = (node_status[node_at_link_start],
                           node_status[node_at_link_end])

    (active_link_ids, ) = np.where(link_is_active(status_at_link_ends))

    return active_link_ids


def in_link_count_per_node(node_at_link_ends, number_of_nodes=None):
    """in_link_count_per_node((node0, node1), number_of_nodes=None)
    Number of links entering nodes.

    Parameters
    ----------
    node0, node1 : sequence of array-like
        Node ID at link start and end.
    number_of_nodes : int, optional
        Number of nodes in the grid

    Returns
    -------
    ndarray :
        Number of links entering nodes.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import in_link_count_per_node
    >>> link_ends = [(0, 3), (1, 4), (2, 5), (3, 6), (4, 7), (5, 8)]
    >>> in_link_count_per_node(zip(*link_ends))
    array([0, 0, 0, 1, 1, 1, 1, 1, 1])
    """
    node_at_link_start, node_at_link_end = _split_link_ends(node_at_link_ends)

    #if len(node_at_link_end) != len(node_at_link_start):
    #    raise ValueError('Link arrays must be the same length')

    return np.bincount(node_at_link_end, minlength=number_of_nodes)


def out_link_count_per_node(node_at_link_ends, number_of_nodes=None):
    """out_link_count_per_node((node0, node1), number_of_nodes=None)
    Number of links leaving nodes.

    Parameters
    ----------
    node0, node1 : sequence of array-like
        Node ID at link start and end.
    number_of_nodes : int, optional
        Number of nodes in the grid

    Returns
    -------
    ndarray :
        Number of links leaving nodes.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import out_link_count_per_node
    >>> out_link_count_per_node(([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7, 8]))
    array([1, 1, 1, 1, 1, 1])
    >>> out_link_count_per_node(([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7, 8]),
    ...     number_of_nodes=9)
    array([1, 1, 1, 1, 1, 1, 0, 0, 0])
    """
    node_at_link_start, node_at_link_end = _split_link_ends(node_at_link_ends)
    if len(node_at_link_end) != len(node_at_link_start):
        raise ValueError('Link arrays must be the same length')
    return np.bincount(node_at_link_start, minlength=number_of_nodes)


def link_count_per_node(node_at_link_ends, number_of_nodes=None):
    """link_count_per_node((node0, node1), number_of_nodes=None)
    Number of links per node.

    Parameters
    ----------
    node0, node1 : sequence of array-like
        Node ID at link start and end.
    number_of_nodes : int, optional
        Number of nodes in the grid

    Returns
    -------
    ndarray :
        Number of links per nodes.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import link_count_per_node
    >>> link_count_per_node(([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7, 8]))
    array([1, 1, 1, 2, 2, 2, 1, 1, 1])
    """
    in_count = in_link_count_per_node(node_at_link_ends)
    out_count = out_link_count_per_node(node_at_link_ends)

    node_count = number_of_nodes or max(len(in_count), len(out_count))

    if len(in_count) < node_count:
        in_count = np.pad(in_count, (0, node_count - len(in_count)),
                          mode='constant')
    if len(out_count) < node_count:
        out_count = np.pad(out_count, (0, node_count - len(out_count)),
                           mode='constant')

    return in_count + out_count


def _sort_links_by_node(node_at_link_ends, link_ids=None, sortby=0):
    sorted_links = np.argsort(node_at_link_ends[sortby])

    if link_ids is not None:
        return np.array(link_ids)[sorted_links]
    else:
        return sorted_links


def in_link_ids_at_node(node_at_link_ends, link_ids=None, number_of_nodes=None):
    """in_link_ids_at_node((node0, node1), number_of_nodes=None)
    Links entering nodes.

    Parameters
    ----------
    node0, node1 : sequence of array-like
        Node ID at link start and end.
    number_of_nodes : int, optional
        Number of nodes in the grid

    Returns
    -------
    tuple :
        Tuple of link id array and offset into link id array.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import in_link_ids_at_node
    >>> (links, count) = in_link_ids_at_node(([0, 1, 2, 3, 4, 5],
    ...                                       [3, 4, 5, 6, 7, 8]))
    >>> links
    array([0, 1, 2, 3, 4, 5])
    >>> count
    array([0, 0, 0, 1, 1, 1, 1, 1, 1])


    >>> (links, count) = in_link_ids_at_node(([0, 1, 2, 3, 4, 5],
    ...                                       [3, 4, 5, 6, 7, 8]),
    ...                                      link_ids=range(1, 7))
    >>> links
    array([1, 2, 3, 4, 5, 6])
    >>> count
    array([0, 0, 0, 1, 1, 1, 1, 1, 1])
    """
    node_at_link_ends = _split_link_ends(node_at_link_ends)

    link_ids = _sort_links_by_node(node_at_link_ends, link_ids=link_ids,
                                   sortby=1)
    links_per_node = in_link_count_per_node(node_at_link_ends,
                                            number_of_nodes=number_of_nodes)
    return link_ids, links_per_node


def out_link_ids_at_node(node_at_link_ends, link_ids=None, number_of_nodes=None):
    """out_link_ids_at_node((node0, node1), number_of_nodes=None)
    Links leaving nodes.

    Parameters
    ----------
    node0, node1 : sequence of array-like
        Node ID at link start and end.
    number_of_nodes : int, optional
        Number of nodes in the grid

    Returns
    -------
    tuple :
        Tuple of link id array and offset into link id array.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import out_link_ids_at_node
    >>> (links, count) = out_link_ids_at_node(
    ...     ([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7, 8]), link_ids=range(-1, 5),
    ...     number_of_nodes=9)
    >>> links
    array([-1,  0,  1,  2,  3,  4])
    >>> count
    array([1, 1, 1, 1, 1, 1, 0, 0, 0])


    >>> (links, count) = out_link_ids_at_node(
    ...     ([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7, 8]), number_of_nodes=9)
    >>> links
    array([0, 1, 2, 3, 4, 5])
    >>> count
    array([1, 1, 1, 1, 1, 1, 0, 0, 0])
    """
    node_at_link_ends = _split_link_ends(node_at_link_ends)

    link_ids = _sort_links_by_node(node_at_link_ends, link_ids=link_ids,
                                   sortby=0)
    links_per_node = out_link_count_per_node(node_at_link_ends,
                                             number_of_nodes=number_of_nodes)
    return link_ids, links_per_node


def link_ids_at_node(node_at_link_ends, number_of_nodes=None):
    """link_ids_at_node((node0, node1), number_of_nodes=None)
    Links entering and leaving nodes.

    Parameters
    ----------
    node0, node1 : sequence of array-like
        Node ID at link start and end.
    number_of_nodes : int, optional
        Number of nodes in the grid

    Returns
    -------
    tuple :
        Tuple of link id array and offset into link id array.

    Examples
    --------
    >>> from landlab.grid.unstructured.links import link_ids_at_node
    >>> (links, count) = link_ids_at_node(
    ...     ([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7, 8]), number_of_nodes=9)
    >>> links
    array([0, 1, 2, 0, 3, 1, 4, 2, 5, 3, 4, 5])
    >>> count
    array([1, 1, 1, 2, 2, 2, 1, 1, 1])
    """
    links_per_node = link_count_per_node(node_at_link_ends,
                                         number_of_nodes=number_of_nodes)

    in_links = JaggedArray(
        *in_link_ids_at_node(node_at_link_ends,
                             number_of_nodes=number_of_nodes))
    out_links = JaggedArray(
        *out_link_ids_at_node(node_at_link_ends,
                              number_of_nodes=number_of_nodes))

    links = np.empty(in_links.size + out_links.size, dtype=int)

    offset = 0
    for node, link_count in enumerate(links_per_node):
        links[offset:offset + link_count] = np.concatenate(
            (in_links.row(node), out_links.row(node)))
        offset += link_count

    return links, links_per_node


class LinkGrid(object):
    def __init__(self, link_ends, number_of_nodes, link_ids=None,
                 node_status=None):
        """Create a grid of links that enter and leave nodes.
        __init__((node0, node1), number_of_nodes=None)

        Parameters
        ----------
        node0, node1 : sequence of array-like
            Node ID at link start and end.
        number_of_nodes : int, optional
            Number of nodes in the grid

        Returns
        -------
        LinkGrid :
            A newly-created grid

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2, 0), (2, 3, 1, 3, 3)], 4)
        >>> lgrid.number_of_links
        5
        >>> lgrid.number_of_nodes
        4
        >>> lgrid.number_of_in_links_at_node(0)
        0
        >>> lgrid.number_of_out_links_at_node(0)
        3
        >>> lgrid.out_link_at_node(0)
        array([0, 2, 4])
        >>> lgrid.nodes_at_link_id(1)
        array([1, 3])

        >>> lgrid = LinkGrid([(0, 1, 0, 2, 0), (2, 3, 1, 3, 3)], 4,
        ...                  link_ids=range(1, 6))
        >>> lgrid.nodes_at_link
        array([[0, 2],
               [1, 3],
               [0, 1],
               [2, 3],
               [0, 3]])
        >>> lgrid.out_link_at_node(0)
        array([1, 3, 5])
        >>> lgrid.nodes_at_link_id(1)
        array([0, 2])
        """
        self._in_link_at_node = JaggedArray(
            *in_link_ids_at_node(link_ends, link_ids=link_ids,
                                 number_of_nodes=number_of_nodes)

        )
        self._out_link_at_node = JaggedArray(
            *out_link_ids_at_node(link_ends, link_ids=link_ids,
                                  number_of_nodes=number_of_nodes)
        )
        self._link_ends = np.array(link_ends)
        if link_ids is not None:
            self._link_id_map = dict(zip(link_ids, xrange(len(link_ids))))
            self._link_ids = link_ids

        self._number_of_links = len(link_ends[0])
        self._number_of_nodes = number_of_nodes

        self._node_status = node_status

    @property
    def number_of_links(self):
        """Number of links in the grid.
        """
        return self._number_of_links

    @property
    def number_of_nodes(self):
        """Number of nodes in the grid.
        """
        return self._number_of_nodes

    def number_of_in_links_at_node(self, node):
        """Number of links entering a node.

        Parameters
        ----------
        node : int
            Node ID

        Returns
        -------
        int :
            Number of links entering the node.

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2), (2, 3, 1, 3)], 4)
        >>> [lgrid.number_of_in_links_at_node(node) for node in xrange(4)]
        [0, 1, 1, 2]
        """
        return self._in_link_at_node.length_of_row(node)

    def number_of_out_links_at_node(self, node):
        """Number of links leaving a node.

        Parameters
        ----------
        node : int
            Node ID

        Returns
        -------
        int :
            Number of links leaving the node.

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2), (2, 3, 1, 3)], 4)
        >>> [lgrid.number_of_out_links_at_node(node) for node in xrange(4)]
        [2, 1, 1, 0]
        """
        return self._out_link_at_node.length_of_row(node)

    def number_of_links_at_node(self, node):
        """Number of links entering and leaving a node.

        Parameters
        ----------
        node : int
            Node ID

        Returns
        -------
        int :
            Number of links entering and leaving the node.

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2), (2, 3, 1, 3)], 4)
        >>> [lgrid.number_of_links_at_node(node) for node in xrange(4)]
        [2, 2, 2, 2]
        """
        return (self.number_of_in_links_at_node(node) +
                self.number_of_out_links_at_node(node))

    @property
    def node_at_link_start(self):
        return self._link_ends[0]

    @property
    def node_at_link_end(self):
        return self._link_ends[1]

    @property
    def nodes_at_link(self):
        return self._link_ends.T

    @property
    def link_id(self):
        try:
            return self._link_ids
        except AttributeError:
            return np.arange(self.number_of_links)

    def nodes_at_link_id(self, link_id):
        try:
            return self.nodes_at_link[self._link_id_map[link_id]]
        except AttributeError:
            return self.nodes_at_link[link_id]


    def in_link_at_node(self, node):
        """Links entering a node.

        Parameters
        ----------
        node : int
            Node ID

        Returns
        -------
        ndarray :
            Links entering the node

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2), (2, 3, 1, 3)], 4)
        >>> lgrid.in_link_at_node(0)
        array([], dtype=int64)
        >>> lgrid.in_link_at_node(3)
        array([1, 3])
        """
        return self._in_link_at_node.row(node)

    def out_link_at_node(self, node):
        """Links leaving a node.

        Parameters
        ----------
        node : int
            Node ID

        Returns
        -------
        ndarray :
            Links leaving the node

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2), (2, 3, 1, 3)], 4)
        >>> lgrid.out_link_at_node(0)
        array([0, 2])
        >>> lgrid.out_link_at_node(3)
        array([], dtype=int64)
        """
        return self._out_link_at_node.row(node)

    def iter_nodes(self):
        """Iterate of the nodes of the grid.

        Returns
        -------
        ndarray :
            Links entering and leaving each node

        Examples
        --------
        >>> from landlab.grid.unstructured.links import LinkGrid
        >>> lgrid = LinkGrid([(0, 1, 0, 2), (2, 3, 1, 3)], 4)
        >>> for link in lgrid.iter_nodes(): link
        array([0, 2])
        array([2, 1])
        array([0, 3])
        array([1, 3])
        """
        for node in xrange(self.number_of_nodes):
            yield np.concatenate((
                self.in_link_at_node(node),
                self.out_link_at_node(node),
            ))

    @property
    def node_status_at_link_start(self):
        return self._node_status[self.node_at_link_start]

    @property
    def node_status_at_link_end(self):
        return self._node_status[self.node_at_link_end]



#! /bin/env python
import os
import xml.dom.minidom


_VALID_ENCODINGS = set(['ascii', 'base64', 'raw'])
_VALID_FORMATS = set(['ascii', 'base64', 'raw', 'appended'])
_VTK_POSSIBLE_PIECE_SECTIONS = [
    'Points', 'Coordinates', 'PointData', 'Cells', 'CellData',
]


class VtkError(Exception):
    pass


class InvalidFormatError(VtkError):
    def __init__(self, format):
        self.name = format

    def __str__(self):
        return '%s: Invalid format' % self.name


class InvalidEncodingError(VtkError):
    def __init__(self, encoding):
        self.name = encoding

    def __str__(self):
        return '%s: Invalid encoder' % self.name


def assemble_vtk_document(element):
    root = element['VTKFile']
    grid = element['Grid']
    piece = element['Piece']

    for section in _VTK_POSSIBLE_PIECE_SECTIONS:
        try:
            piece.appendChild(element[section])
        except (KeyError, EmptyElementError):
            pass
    grid.appendChild(piece)
    root.appendChild(grid)
  
    try:
        root.appendChild(element['AppendedData'])
    except KeyError:
        pass
  
    return root


def assert_format_is_valid(format_string):
    if format not in _VALID_FORMATS:
        raise InvalidFormatError(format)


def assert_encoding_is_valid(encoding_string):
    if encoding not in _VALID_ENCODINGS:
        raise InvalidEncodingError(encoding)
  
  
class VtkWriter(xml.dom.minidom.Document):
    def __init__(self, **kwds):
        self._format = kwds.pop('format', 'ascii')
        self._encoding = kwds.pop('encoding', 'ascii')

        assert_format_is_valid(self.format)
        assert_encoding_is_valid(self.encoding)

        if format == 'ascii':
            encoding = 'ascii'

        if self.format == 'appended':
            self._data = VtkAppendedDataElement('', encoding=self.encoding)
        else:
            self._data = None
  
        xml.dom.minidom.Document.__init__(self)

    @property
    def format(self):
        return self._format

    @property
    def encoding(self):
        return self._encoding

    @property
    def data(self):
        return self._data

    def construct_field_elements(self, field):
        raise NotImplementedError()

    def write(self, path, field):
        self.unlink()
        elements = self.construct_field_elements(field)

        if self.data is not None:
            element['AppendedData'] = self.data

        self.appendChild(assemble_vtk_document(elements))
        self.to_xml(path)
  
    def to_xml(self, path):
        with open(path, 'w') as xml_file:
            xml_file.write(self.toprettyxml())


def assemble_vtk_elements(element):
    root = element['VTKFile']
    grid = element['Grid']
    piece = element['Piece']
    for section in _VTK_FILE_SECTIONS:
        try:
            piece.appendChild(element[section])
        except (KeyError, EmptyElementError):
            pass
    grid.appendChild(piece)
    root.appendChild(grid)

    try:
        root.appendChild(element['AppendedData'])
    except KeyError:
        pass

    return root


class VTKDatabase(VtkWriter):
    def write (self, path, **kwargs):
        (base, file) = os.path.split(path)
        (root, ext) = os.path.splitext(file)

        try:
            next_file = '%s_%04d%s' % (root, self._count, ext)
        except NameError:
            self._count = 0
            next_file = '%s_%04d%s' % (root, self._count, ext)

        VtkWriter.write(self, os.path.join (base, next_file), **kwargs)

        self._count += 1


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
hex_lattice_tectonicizer.py

Models discrete normal-fault offset on a 2D hex lattice with a rectangular
shape and with one orientation of the nodes being vertical.

The intention here is to use a particle (LCA) model to represent the evolution
of a 2D hillslope, with the hex_lattice_tectonicizer serving to shift the nodes
either upward (simple vertical uplift relative to baselevel), or up and
sideways (representing motion on a fault plane).

Created on Mon Nov 17 08:01:49 2014

@author: gtucker
"""

from landlab import HexModelGrid
from numpy import amax, zeros, arange, array
from pylab import figure, draw, show

_DEFAULT_NUM_ROWS = 5
_DEFAULT_NUM_COLS = 5
_TAN60 = 1.732


class HexLatticeTectonicizer(object):
    """Base class from which classes to represent particular baselevel/fault
    geometries are derived.
    """
    def __init__(self, grid=None, node_state=None, propid=None, prop_data=None, prop_reset_value=None):
        
        # If needed, create grid
        if grid is None:
            num_rows = _DEFAULT_NUM_ROWS
            num_cols = _DEFAULT_NUM_COLS
            self.grid = HexModelGrid(num_rows, num_cols, dx=1.0,
                                     orientation='vertical',
                                     shape='rect', reorient_links=True)
        else:
            # Make sure caller passed the right type of grid            
            assert (grid.orientation=='vertical'), \
                   'Grid must have vertical orientation'
                   
            # Keep a reference to the grid
            self.grid = grid
            
        # If needed, create node-state grid
        if node_state is None:
            self.node_state = self.grid.add_zeros('node', 'node_state_map')
        else:
            #print 'setting node state'
            self.node_state = node_state
            
        # Remember the # of rows and cols
        self.nr = self.grid.number_of_node_rows
        self.nc = self.grid.number_of_node_columns
        
        # propid should be either a reference to a CA model's "property id" 
        # array, or None
        self.propid = propid
        self.prop_data = prop_data
        self.prop_reset_value = prop_reset_value
        

class LatticeNormalFault(HexLatticeTectonicizer):
    """Represents a 60 degree, left-dipping normal fault, and handles discrete
    offsets for a hex grid that has vertical columns and a rectangular shape.
    

    """
    def __init__(self, fault_x_intercept=0.0, grid=None, node_state=None, propid=None, prop_data=None, prop_reset_value=None):
        """        
        Examples
        --------
        >>> import numpy as np
        >>> from landlab import HexModelGrid
        >>> from landlab.components.cellular_automata.boundaries.hex_lattice_tectonicizer import LatticeNormalFault

        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        1
        >>> lnf.num_fw_rows
        array([0, 1, 3, 4, 5])
        >>> lnf.incoming_node
        array([ 5, 10, 11, 15])
        >>> lnf.outgoing_node
        array([22, 23, 24, 18])
        """
        # Do the base class init
        super(LatticeNormalFault, self).__init__(grid, node_state, propid, prop_data, prop_reset_value)

        # Set up data structures:
        #   Make sure the footwall location is such that the fault actually
        #   cuts across the grid. This means the x intercept has to be, at
        #   the very least, no smaller than the biggest x-coordinate, and if
        #   there is an even number of columns, it must be smaller than that
        #   number minus 1/tangent 60 degrees (0.57735)
        assert (fault_x_intercept<amax(self.grid.node_x)-0.57735)
        
        #   Figure out which nodes are and are not within the footwall
        in_footwall = (self.grid.node_y<_TAN60*(self.grid.node_x-fault_x_intercept))
        
        #   Find the first of the bottom-row nodes that lies in the footwall
        self.first_fw_col = 0
        n = 0
        while not in_footwall[n]:
            n += self.nr
            self.first_fw_col += 1
            
        #   Remember the number of footwall rows in each column
        self.num_fw_rows = zeros(self.nc, dtype=int)
        for c in range(self.nc):
            current_row = 0
            while current_row<self.nr and in_footwall[current_row+c*self.nr]:
                self.num_fw_rows[c] += 1
                current_row += 1
                
        # If we're handling properties and property IDs, we need to do some setup
        if self.propid is not None:
            
            # We want to remember the node IDs of two sets of nodes: those 
            # whose contents will vanish off the right-hand side (and possibly
            # the top) of the grid with each offset step, and those that gain
            # new ("rock") contents with each such step.
            #
            # First, let's find the latter set, which we'll call 
            # "incoming_node" because material is flowing into these nodes from
            # below. To do this, we'll go column-by-column, starting with
            # the first column that has nodes in the footwall, and going to
            # the next-to-last column. Even-numbered columns have *two*
            # incoming nodes at their base (if the footwall reaches that high),
            # whereas odd-numbered columns have only one. Note that we'll start
            # with a list (so we can append), and then convert to an array
            # that belongs to this class.
            incoming_node_list = []
            for c in range(self.first_fw_col, self.nc-1):
                
                # This little loop appends to the incoming node list: 
                # either just the bottom node in this column (ID=self.nr*c)
                # or that plus the node above it (ID=self.nr*c+1). Note that
                # 2-c%s evaluates to 2 for even-numbered columns and 1 for 
                # odd-numbered columns. So the loop is either 1 or 2 iterations.
                for node_id in range(self.nr*c, self.nr*c+(2-c%2)):
                    incoming_node_list.append(node_id)
                    
            # Convert to a numpy array that belongs to the class (so we can
            # use it in the do_offset() method).
            self.incoming_node = array(incoming_node_list, dtype=int)
            
            # Next, find the IDs the "outgoing" nodes. There will always be
            # some of these along the right side of the grid. Depending on the
            # height of the grid and the position of the fault, there may or
            # may not also be some along the top.
            #
            # Which of the nodes on the right side outgoing? The lowermost one
            # won't be. If the right-most column is even-numbered, the next one 
            # up won't be either. So call the ID of the first potential outgoing 
            # node onthe right side the "base_id". Then we also have the "top_id",
            # which is the ID of *either* the top-most node of the column
            # (ID=# of nodes in grid-1), *or* the top footwall node.
            #
            # The next line translates as: Take the ID of the bottom node in 
            # the right-most column (nr x (nc-1)) and add to it either 1
            # (if the last column is odd-numbered) or 2 (if even-numbered).
            # Example: in a 5x5 grid, it is node ID 22 (two up from the base)
            base_id = self.nr*(self.nc-1)+(1+self.nc%2)
            # 
            # The next line finds the ID of the top outgoing node on the right
            # side. Here's how it works. Find the ID of the topmost node in
            # the column: that's the # of nodes in the grid minus one. 
            # Example: in a 5x5 grid, that's node number 24.
            # 
            # Then find the ID of the topmost node IN THE FOOTWALL. To get this, 
            # we start with the ID at the top of the column to the left, and we start  
            # we start number of footwall nodes in the right-most column.
            # Example: in a 5x5 grid, the top of the next-to-rightmost column
            # is node 19. If the fault position is x=0.0, there will be 5
            # footwall nodes here. 19+5 = 24 (meaning the top of the column
            # and the top of the footwall happen to the same)
            top_id = min(self.nr*self.nc-1, (self.nr*(self.nc-1)-1)+self.num_fw_rows[self.nc-1])
            
            # Having found the top and the base, we now append all these to
            # the list of outgoing nodes.
            outgoing_node_list = []
            for node_id in range(base_id, top_id+1):
                outgoing_node_list.append(node_id)
                
            # Now, at this point, we might have found all the outgoing nodes.
            # Or there might be some more along the top edge of the grid. The
            # latter will be true if the number of outgoing nodes found so far
            # is less than the number there should be (same as # of incoming).
            #
            # start with the next-to-rightmost column; if we need to, we'll
            # work right-to-left from column to column
            col = self.nc-2
            # If the following is true, we still have more nodes to add along
            # the top
            while len(outgoing_node_list)<len(self.incoming_node):
                
                # Add the top-most footwall node in this column
                outgoing_node_list.append(self.nr*col+self.num_fw_rows[col]-1)
                
                # If we're on an odd-numbered column, and not already full,
                # add the next one down too
                if col%2==1 and len(outgoing_node_list)<len(self.incoming_node):
                    outgoing_node_list.append(self.nr*col+self.num_fw_rows[col]-2)
                    
            # Finally, convert the outgoing node list to an array stored in this
            # object
            self.outgoing_node = array(outgoing_node_list, dtype=int)
            
        
    def do_offset(self, rock_state=1):
        """Applies 60-degree normal-fault offset to a hexagonal grid with
        vertical node orientation and rectangular arrangement of nodes.
        
        Parameters
        ----------
        rock_state : int
            State code to apply to new cells introduced along bottom row.
            
        Examples
        --------
        >>> import numpy as np
        >>> from landlab.components.cellular_automata.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
        >>> from landlab import HexModelGrid

        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.do_offset()
        >>> lnf.propid
        array([ 0,  1,  2,  3,  4, 22,  6,  7,  8,  9, 23, 24,  5, 13, 14, 18, 10,
               11, 12, 19, 20, 21, 15, 16, 17])
        """

        # If we need to shift the property ID numbers, we'll first need to
        # record the property IDs in those nodes that are about to "shift off
        # the grid" (or rather, their contents will shift) due to tectonic
        # motion. We'll call these "nodes to replace".
        if self.propid is not None:
            
            # Now, remember the property IDs of the nodes along the right side
            # and possibly top that are about to shift off the grid
            propids_for_incoming_nodes = self.propid[self.outgoing_node]
            #print 'pid for new base:',propids_for_incoming_nodes

        # We go column-by-column, starting from the right side
        for c in range(self.grid.number_of_node_columns-1, self.first_fw_col-1, -1):
            
            # Odd-numbered rows are shifted up in the hexagonal, vertically
            # oriented lattice
            row_offset = 2-c%2
            
            # Number of base nodes in the footwall in this column. Either 1 or 2.
            n_base_nodes = min(self.num_fw_rows[c], row_offset)
            
            # ID of the bottom footwall node
            bottom_node = c*self.grid.number_of_node_rows
            
            # The bottom 1 or 2 nodes in this column are set to rock
            self.node_state[bottom_node:(bottom_node+n_base_nodes)] = rock_state
            
            # "indices" here contains the array indices of those nodes in this
            # column that are to be replaced by the ones in the column to the
            # left and down one or two nodes. We do this replacement if indices
            # contains any data.
            indices = arange(n_base_nodes, self.num_fw_rows[c], dtype=int)+bottom_node
            if len(indices)>0:
                self.node_state[indices] = self.node_state[indices-(self.nr+row_offset)]
                if self.propid is not None:
                    #print 'replacing nodes',indices,'with',indices-(self.nr+row_offset)
                    self.propid[indices] = self.propid[indices-(self.nr+row_offset)]

        if self.propid is not None:
            self.propid[self.incoming_node] = propids_for_incoming_nodes
            self.prop_data[self.propid[self.incoming_node]] = self.prop_reset_value
            #print 'pid after:',self.propid
            #print 'propdata after:',self.prop_data

        if self.first_fw_col==0:
            self.node_state[:self.n_footwall_rows[0]] = rock_state


class LatticeUplifter(HexLatticeTectonicizer):
    """Handles vertical uplift of interior (not edges) for a hexagonal lattice
    with vertical node orientation and rectangular node arrangement.
    """
    def __init__(self, grid=None, node_state=None, propid=None, prop_data=None, prop_reset_value=None):
        
        # Do the base class init
        super(LatticeUplifter, self).__init__(grid, node_state, propid, prop_data, prop_reset_value)

        self.base_row_nodes = arange(self.nr, self.nr*(self.nc-1), self.nr)
        #print 'LU INIT HERE******************'
        if self.propid is not None:
            self.top_row_nodes = self.base_row_nodes+(self.nr-1)
            #print 'top:',self.top_row_nodes
            #print 'base:',self.base_row_nodes
    
    def uplift_interior_nodes(self, rock_state=1):

        #print 'in uin, ns is'
        #print self.node_state 
        #print 'and propid before is'
        #print self.propid
        #print 'and prop before is'
        #print self.prop_data[self.propid]
        
        # Shift the node states up by one row
        for r in range(self.nr-1, 0, -1):
            self.node_state[self.base_row_nodes+r] = \
                    self.node_state[self.base_row_nodes+(r-1)]
        self.node_state[self.base_row_nodes] = rock_state
        
        # If propid (property ID or index) is defined, shift that too.
        #print 'uid1: propid 11-12=',self.propid[11:13],self.prop_data[self.propid[11:13]]
        if self.propid is not None:
            top_row_propid = self.propid[self.top_row_nodes]
            for r in range(self.nr-1, 0, -1):
                self.propid[self.base_row_nodes+r] = self.propid[self.base_row_nodes+(r-1)]
            #print 'uid2: propid 11-12=',self.propid[11:13],self.prop_data[self.propid[11:13]]
            self.propid[self.base_row_nodes] = top_row_propid
            #print 'uid3: propid 11-12=',self.propid[11:13],self.prop_data[self.propid[11:13]]
            self.prop_data[self.propid[self.base_row_nodes]] = self.prop_reset_value
            #print 'uid4: propid 11-12=',self.propid[11:13],self.prop_data[self.propid[11:13]]
            #print 'in UIN, pid is'
            #print self.propid
            #print self.prop_data[self.propid]


def test_create_lnf(nr, nc):
    
    pid = arange(nr*nc, dtype=int)
    ns = arange(nr*nc, dtype=int)
    pdata = arange(nr*nc)
    grid = HexModelGrid(nr, nc, 1.0, orientation='vertical', shape='rect', reorient_links=True)
    lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    return lnf


def main():
    """The main function is used just to do some unit tests.
    
    Examples
    --------
    >>> from landlab.components.cellular_automata.boundaries.hex_lattice_tectonicizer import test_create_lnf
    >>> lnf = test_create_lnf(4, 4)
    >>> lnf.incoming_node
    array([4, 8, 9])
    >>> lnf.outgoing_node
    array([13, 14, 15])
    >>> lnf.do_offset()
    >>> lnf.propid
    array([ 0,  1,  2,  3, 13,  5,  6,  7, 14, 15,  4, 11, 12,  8,  9, 10])
    >>> lnf = test_create_lnf(4, 5)
    >>> lnf.incoming_node
    array([ 4,  8,  9, 12])
    >>> lnf.outgoing_node
    array([18, 19, 15, 14])
    >>> lnf.do_offset()
    >>> lnf.propid
    array([ 0,  1,  2,  3, 18,  5,  6,  7, 19, 15,  4, 11, 14,  8,  9, 10, 16,
           17, 12, 13])


    """
    pid = arange(_DEFAULT_NUM_ROWS*_DEFAULT_NUM_COLS, dtype=int)
    pdata = arange(_DEFAULT_NUM_ROWS*_DEFAULT_NUM_COLS)
    lnf = LatticeNormalFault(propid=pid, prop_data=pdata, prop_reset_value=0.0)

    for i in range(3):
        lnf.do_offset()
    lnf.grid.hexplot(lnf.node_state)
    show()
    
    lu = LatticeUplifter()
    lu.uplift_interior_nodes()
    figure(2)
    for i in range(2):
        lu.uplift_interior_nodes()
        lu.grid.hexplot(lu.node_state)
    
if __name__=='__main__':
    import doctest
    doctest.testmod()
    main()



#! /usr/env/python
"""

Classes for Landlab that deal with setting boundary conditions if running
on a DEM.  This takes advantage of methods in the grid, but combines them
to make it easier for a user.

Last updated NG 8/2013

"""

from landlab import RasterModelGrid
import numpy as np
import pylab

class WatershedBoundaryConditions(object):
    """
    If using a DEM of a watershed, this class will properly 
    set the boundary conditions.
    """
    
    def __init__(self):
        """
        what does this need?
        """
        
    def set_bc_node_coords(self, mg, node_data, nodata_value, outlet_row, outlet_column):
        """
        Sets the boundary conditions for a watershed.
        Assumes that outlet is already known.
        
        This must be passed the grid, node_data and nodata_value, 
        and the values of the outlet_row and outlet_column.
        """
        
        #set no data nodes to inactive boundaries
        mg.set_nodata_nodes_to_inactive(node_data, nodata_value)
        
        #find the id of the outlet node
        outlet_node = mg.grid_coords_to_node_id(outlet_row, outlet_column)
        #set the boundary condition (fixed value) at the outlet_node
        mg.set_fixed_value_boundaries(outlet_node)
    
    def set_bc_node_id(self, mg, node_data, nodata_value, outlet_node):
        """
        Sets the boundary conditions for a watershed.
        Assumes that outlet is already known.
        
        This must be passed the grid, node_data and nodata_value, 
        and the id of the outlet node.
        """
        
        #set no data nodes to inactive boundaries
        mg.set_nodata_nodes_to_inactive(node_data, nodata_value)
        
        #set the boundary condition (fixed value) at the outlet_node        
        mg.set_fixed_value_boundaries(outlet_node)
        
    def set_bc_find_outlet(self, mg, node_data, nodata_value):
        """
        Finds the node adjacent to a boundary node with the smallest value.
        This node is set as the outlet.
        
        This must be passed the grid, node_data and nodata_value.
        """
        #for this to be a watershed, need to make sure that there is a ring
        #of no data values around the outside of the watershed, barring the 
        #outlet location.  So enforce that all outer nodes 
        #are inactive boundaries now, then set the outlet location later.
        #By enforcing the ring of closed values first, then fixing the outlet
        #later, it should be OK if the outlet is on the outer ring.
        mg.set_inactive_boundaries(True, True, True, True)
        
        #set no data nodes to inactive boundaries
        #this may be redundant, but must do in case there are no data
        #values that are not on the outer boundary
        mg.set_nodata_nodes_to_inactive(node_data, nodata_value)
        
        #This method works well if the watershed topography is already 
        #established.  If it's not, then this is an ineffiient method, but 
        #seems likely that one would only call this if the watershed
        #topography was already established.
        
        #ng this could maybe be generalized (?) so you don't need this if/else
        #first case is if nodata_value is minimum value
        if (min(node_data) == nodata_value):
            #min value is nodata_value, so need to find values
            #that are not no_data
            
            #locs is a list that contains locations where
            #node data is greater than the nodata value
            locs=list(np.where(node_data>nodata_value)[0])
            
            #now find minimum of the data values
            min_val=np.min(node_data[locs])
            
            #now find where minimum values are
            min_locs=list(np.where(node_data==min_val)[0])
  
            #check all the locations with the minimum value to see if one
            #is adjacent to a boundary location.  If so, that will be the
            #watershed outlet.  If none of these points qualify, then 
            #increase the minimum value and check again.  Keep checking
            #until a point next to the boundary is found.
            #
            #NG I think the only way this would become an infinite loop
            #is if there are no interior nodes.          
            not_found=True          
            while not_found:
                #now check these locations to see if any are next to 
                #a boundary node          
                local_not_found=True
                i=0
                while (i<len(min_locs) and local_not_found):
                    if mg.has_boundary_neighbor(min_locs[i]):
                        local_not_found = False
                        #outlet_loc contains the index of the outlet location
                        #in the node_data array
                        outlet_loc=min_locs[i]
                    else:
                        i += 1
            
                #checked all of the min vals, (so done with inner while)
                #and none of the min values were outlet candidates
                if local_not_found:
                    #need to find the next largest minimum value
                    #first find the locations of all values greater
                    #than the old minimum
                    #not done with outer while
                    locs=list(np.where(node_data>min_val)[0])
                    #now find new minimum of these values
                    min_val=np.min(node_data[locs])
                    min_locs=list(np.where(node_data==min_val)[0])
                else:
                    #if locally found, it is also globally found
                    #so done with outer while
                    not_found = False   
        else:
            #entering this else b/c nodata_value is not the minimum value
            #can I generalize this?
            #will the no data value ever be anything other than the minimum
            #value?
            #this code is exactly the same as above, just have to check that
            #every min value is not the no data value.
            min_val=np.min(node_data)
            if (min_val == nodata_value):
                #found place where no data value is
                #find locations that have values greater than
                #the no data value, then find the min val of these locations
                locs=list(np.where(node_data>nodata_value)[0])
                min_val = np.min(node_data[locs])
            
            #now find where minimum values are
            min_locs=list(np.where(node_data==min_val)[0])
            #check all the locations with the minimum value to see if one
            #is adjacent to a boundary location.  If so, that will be the
            #watershed outlet.  If none of these points qualify, then 
            #increase the minimum value and check again.  Keep checking
            #until a point next to the boundary is found.
            #
            #NG I think the only way this would become an infinite loop
            #is if there are no interior nodes.          
            not_found=True          
            while not_found:
                #now check these locations to see if any are next to 
                #a boundary node
            
                local_not_found=True
                i=0
                while (i<len(min_locs) and local_not_found):
                    if mg.has_boundary_neighbor(min_locs[i]):
                        local_not_found = False
                        #outlet_loc contains the index of the outlet location
                        #in the node_data array
                        outlet_loc=min_locs[i]
                    else:
                        i += 1
            
                #checked all of the min vals, (so done with inner while)
                #and none of the min values were outlet candidates
                if local_not_found:
                    #need to find the next largest minimum value
                    #first find the locations of all values greater
                    #than the old minimum
                    #not done with outer while
                    locs=list(np.where(node_data>min_val)[0])
                    #now find new minimum of these values
                    min_val=np.min(node_data[locs])
                    if (min_val == nodata_value):
                        #found place where no data value is
                        #find locations that have values greater than
                        #the no data value, then find the min val of these locations
                        locs=list(np.where(node_data>nodata_value)[0])
                        min_val = np.min(node_data[locs])
                    #now find where minimum values are
                    min_locs=list(np.where(node_data==min_val)[0])
                else:
                    #if locally found, it is also globally found
                    #so done with outer while
                    not_found = False   
                    
        #set outlet boundary condition
        mg.set_fixed_value_boundaries(outlet_loc)
        #x=mg.node_x[outlet_loc]
        #y=node_y[outlet_loc]
        #print "outlet_loc ", outlet_loc," x ",x," y ",y 
        return outlet_loc
        
        



import os

try:
    import netCDF4 as nc4
except ImportError:
    import warnings
    warnings.warn('Unable to import netCDF4.', ImportWarning)
    WITH_NETCDF4 = False
else:
    WITH_NETCDF4 = True

NETCDF4_EXAMPLE_FILE = os.path.join(os.path.dirname(__file__), 'tests', 'data',
                                    'test-netcdf4.nc')
NETCDF3_64BIT_EXAMPLE_FILE = os.path.join(os.path.dirname(__file__), 'tests',
                                          'data', 'test-netcdf3-64bit.nc')


from .read import read_netcdf
from .write import write_netcdf
from .errors import NotRasterGridError


__all__ = ['read_netcdf', 'write_netcdf', 'NotRasterGridError',
           'WITH_NETCDF4', 'NETCDF4_EXAMPLE_FILE',
           'NETCDF3_64BIT_EXAMPLE_FILE']



import numpy as np


from . import nodes
from ..base import CORE_NODE, CLOSED_BOUNDARY
from ..unstructured.links import LinkGrid


def shape_of_vertical_links(shape):
    """Shape of vertical link grid.

    Number of rows and columns of *vertical* links that connect nodes in a
    structured grid of quadrilaterals.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple of int :
        Shape of the vertical links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import shape_of_vertical_links
    >>> shape_of_vertical_links((3, 4))
    (2, 4)
    """
    return (shape[0] - 1,  shape[1])


def shape_of_horizontal_links(shape):
    """Shape of horizontal link grid.

    Number of rows and columns of *horizontal* links that connect nodes in a
    structured grid of quadrilaterals.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple of int :
        Shape of the horizontal links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import shape_of_horizontal_links
    >>> shape_of_horizontal_links((3, 4))
    (3, 3)
    """
    return (shape[0],  shape[1] - 1)


def number_of_vertical_links(shape):
    """Number of vertical links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of vertical links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_vertical_links
    >>> number_of_vertical_links((3, 4))
    8
    """
    return np.prod(shape_of_vertical_links(shape))


def number_of_horizontal_links(shape):
    """Number of horizontal links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of horizontal links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_horizontal_links
    >>> number_of_horizontal_links((3, 4))
    9
    """
    return np.prod(shape_of_horizontal_links(shape))


def number_of_links(shape):
    """Number of links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_links
    >>> number_of_links((3, 4))
    17
    """
    return number_of_vertical_links(shape) + number_of_horizontal_links(shape)


def vertical_link_ids(shape):
    """IDs of vertical links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        Array of link IDs.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import vertical_link_ids
    >>> vertical_link_ids((3, 4))
    array([[0, 1, 2, 3],
           [4, 5, 6, 7]])
    """
    link_ids = np.arange(number_of_vertical_links(shape))
    return link_ids.reshape(shape_of_vertical_links(shape))


def horizontal_link_ids(shape):
    """IDs of horizontal links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        Array of link IDs.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import horizontal_link_ids
    >>> horizontal_link_ids((3, 4))
    array([[ 8,  9, 10],
           [11, 12, 13],
           [14, 15, 16]])
    """
    link_ids = (np.arange(number_of_horizontal_links(shape)) +
                number_of_vertical_links(shape))
    return link_ids.reshape(shape_of_horizontal_links(shape))


def number_of_links_per_node(shape):
    """Number of links touching each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Array of number of links per node.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import (number_of_links_per_node,
    ...                                            number_of_in_links_per_node,
    ...                                            number_of_out_links_per_node)
    >>> number_of_links_per_node((3, 4))
    array([[2, 3, 3, 2],
           [3, 4, 4, 3],
           [2, 3, 3, 2]])
    >>> (number_of_in_links_per_node((3, 4)) +
    ...  number_of_out_links_per_node((3, 4)))
    array([[2, 3, 3, 2],
           [3, 4, 4, 3],
           [2, 3, 3, 2]])
    """
    link_count = np.empty(shape, np.int)
    link_count[1:-1, 1:-1] = 4
    link_count[(0, -1), 1:-1] = 3
    link_count[1:-1, (0, -1)] = 3
    link_count[(0, 0, -1, -1), (0, -1, 0, -1)] = 2
    return link_count


def number_of_in_links_per_node(shape):
    """Number of links entering each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Array of number of in-links per node.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_in_links_per_node
    >>> number_of_in_links_per_node((3, 4))
    array([[0, 1, 1, 1],
           [1, 2, 2, 2],
           [1, 2, 2, 2]])
    """
    link_count = np.empty(shape, np.int)
    link_count[1:, 1:] = 2
    link_count[0, 0] = 0
    link_count[0, 1:] = 1
    link_count[1:, 0] = 1
    return link_count


def number_of_out_links_per_node(shape):
    """Number of links leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Array of number of out-links per node.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_out_links_per_node
    >>> number_of_out_links_per_node((3, 4))
    array([[2, 2, 2, 1],
           [2, 2, 2, 1],
           [1, 1, 1, 0]])
    """
    link_count = np.empty(shape, np.int)
    link_count[:-1, :-1] = 2
    link_count[-1, -1] = 0
    link_count[-1, :-1] = 1
    link_count[:-1, -1] = 1
    return link_count


def _node_out_link_ids(shape):
    """Link IDs for links leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import _node_out_link_ids
    >>> (vert, horiz) = _node_out_link_ids((3, 4))
    >>> vert
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [-1, -1, -1, -1]])
    >>> horiz
    array([[ 8,  9, 10, -1],
           [11, 12, 13, -1],
           [14, 15, 16, -1]])
    """
    node_horizontal_link_ids = np.empty(shape, np.int)
    node_horizontal_link_ids[:, :-1] = horizontal_link_ids(shape)
    node_horizontal_link_ids[:, -1] = -1

    node_vertical_link_ids = np.empty(shape, np.int)
    node_vertical_link_ids[:-1, :] = vertical_link_ids(shape)
    node_vertical_link_ids[-1, :] = -1

    return node_vertical_link_ids, node_horizontal_link_ids


def _node_in_link_ids(shape):
    """Link IDs for links entering each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import _node_in_link_ids
    >>> (vert, horiz) = _node_in_link_ids((3, 4))
    >>> vert
    array([[-1, -1, -1, -1],
           [ 0,  1,  2,  3],
           [ 4,  5,  6,  7]])
    >>> horiz
    array([[-1,  8,  9, 10],
           [-1, 11, 12, 13],
           [-1, 14, 15, 16]])
    """
    node_horizontal_link_ids = np.empty(shape, np.int)
    node_horizontal_link_ids[:, 1:] = horizontal_link_ids(shape)
    node_horizontal_link_ids[:, 0] = -1

    node_vertical_link_ids = np.empty(shape, np.int)
    node_vertical_link_ids[1:, :] = vertical_link_ids(shape)
    node_vertical_link_ids[0, :] = -1

    return node_vertical_link_ids, node_horizontal_link_ids


def node_in_link_ids(shape):
    """Link IDs for links entering each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_in_link_ids
    >>> (links, offset) = node_in_link_ids((3, 4))
    >>> links
    array([ 8,  9, 10,  0,  1, 11,  2, 12,  3, 13,  4,  5, 14,  6, 15,  7, 16])
    >>> offset
    array([ 0,  0,  1,  2,  3,  4,  6,  8, 10, 11, 13, 15, 17])

    The links entering the 1st, 5th, and last node,

    >>> for link in [0, 4, 11]: links[offset[link]:offset[link + 1]]
    array([], dtype=int64)
    array([0])
    array([ 7, 16])
    """
    (in_vert, in_horiz) = _node_in_link_ids(shape)
    node_link_ids = np.vstack((in_vert.flat, in_horiz.flat)).T
    #offset = np.cumsum(number_of_in_links_per_node(shape))

    offset = np.empty(nodes.number_of_nodes(shape) + 1, dtype=int)
    np.cumsum(number_of_in_links_per_node(shape), out=offset[1:])
    offset[0] = 0
    return node_link_ids[node_link_ids >= 0], offset


def node_out_link_ids(shape):
    """Link IDs for links leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_out_link_ids
    >>> (links, offset) = node_out_link_ids((3, 4))
    >>> links
    array([ 0,  8,  1,  9,  2, 10,  3,  4, 11,  5, 12,  6, 13,  7, 14, 15, 16])
    >>> offset
    array([ 0,  2,  4,  6,  7,  9, 11, 13, 14, 15, 16, 17, 17])

    The links leaving the 1st, 8th, and last node,

    >>> for link in [0, 7, 11]: links[offset[link]:offset[link + 1]]
    array([0, 8])
    array([7])
    array([], dtype=int64)
    """
    (out_vert, out_horiz) = _node_out_link_ids(shape)
    node_link_ids = np.vstack((out_vert.flat, out_horiz.flat)).T
    offset = np.empty(nodes.number_of_nodes(shape) + 1, dtype=int)
    np.cumsum(number_of_out_links_per_node(shape), out=offset[1:])
    offset[0] = 0
    return node_link_ids[node_link_ids >= 0], offset


def node_link_ids(shape):
    """Link IDs for links entering and leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs and offsets into link array.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_link_ids
    >>> (links, offset) = node_link_ids((3, 4))
    >>> links
    array([ 0,  8,  8,  1,  9,  9,  2, 10, 10,  3,  0,  4, 11,  1, 11,  5, 12,
            2, 12,  6, 13,  3, 13,  7,  4, 14,  5, 14, 15,  6, 15, 16,  7, 16])
    >>> offset
    array([ 0,  2,  5,  8, 10, 13, 17, 21, 24, 26, 29, 32, 34])

    The links attached to node 0

    >>> links[offset[0]:offset[1]]
    array([0, 8])

    The links attached to node 5

    >>> links[offset[5]:offset[6]]
    array([ 1, 11,  5, 12])
    """
    (in_vert, in_horiz) = _node_in_link_ids(shape)
    (out_vert, out_horiz) = _node_out_link_ids(shape)
    node_link_ids = np.vstack((in_vert.flat, in_horiz.flat, out_vert.flat, out_horiz.flat)).T

    offset = np.empty(nodes.number_of_nodes(shape) + 1, dtype=int)
    np.cumsum(number_of_links_per_node(shape), out=offset[1:])
    offset[0] = 0

    return node_link_ids[node_link_ids >= 0], offset


def node_id_at_link_start(shape):
    """Node ID at start of links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Node IDs at start of links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_id_at_link_start
    >>> node_id_at_link_start((3, 4))
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  0,  1,  2,  4,  5,  6,  8,  9, 10])
    """
    all_node_ids = nodes.node_ids(shape)
    return np.concatenate((all_node_ids[:-1, :].flat, all_node_ids[:, :-1].flat))


def node_id_at_link_end(shape):
    """Node ID at end of links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Node IDs at end of links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_id_at_link_end
    >>> node_id_at_link_end((3, 4))
    array([ 4,  5,  6,  7,  8,  9, 10, 11,  1,  2,  3,  5,  6,  7,  9, 10, 11])
    """
    all_node_ids = nodes.node_ids(shape)
    return np.concatenate((all_node_ids[1:, :].flat, all_node_ids[:, 1:].flat))


def is_active_link(shape, node_status):
    """IDs of active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : array_link
        Status of nodes in grid.

    Returns
    -------
    ndarray :
        Links IDs at the active links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import status_with_perimeter_as_boundary
    >>> from landlab.grid.structured_quad.links import is_active_link
    >>> status = status_with_perimeter_as_boundary((3, 4))
    >>> is_active_link((3, 4), status)
    array([False,  True,  True, False, False,  True,  True, False, False,
           False, False,  True,  True,  True, False, False, False], dtype=bool)
    """
    if np.prod(shape) != node_status.size:
        raise ValueError('node status array does not match size of grid '
                         '(%d != %d)' % (np.prod(shape), len(node_status)))

    status_at_link_start = node_status.flat[node_id_at_link_start(shape)]
    status_at_link_end = node_status.flat[node_id_at_link_end(shape)]

    return (((status_at_link_start == CORE_NODE) &
             ~ (status_at_link_end == CLOSED_BOUNDARY)) |
            ((status_at_link_end == CORE_NODE) &
             ~ (status_at_link_start == CLOSED_BOUNDARY)))


def active_link_ids(shape, node_status):
    """IDs of active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : array_link
        Status of nodes in grid.

    Returns
    -------
    ndarray :
        Links IDs at the active links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import status_with_perimeter_as_boundary
    >>> from landlab.grid.structured_quad.links import active_link_ids
    >>> status = status_with_perimeter_as_boundary((3, 4))
    >>> active_link_ids((3, 4), status)
    array([ 1,  2,  5,  6, 11, 12, 13])
    """
    return np.where(is_active_link(shape, node_status))[0]


class StructuredQuadLinkGrid(LinkGrid):
    def __init__(self, shape):
        link_ends = (node_id_at_link_start(shape), node_id_at_link_end(shape))
        number_of_nodes = np.prod(shape)
        LinkGrid.__init__(self, link_ends, number_of_nodes)



#! /usr/env/python
"""
diffusion2Dtest.py: Tests ModelGrid() class by implementing a 2D
diffusion code.

version 3 same as 2 but uses source as a dvector

GT, July 2010
"""
import numpy as np
from pylab import plot, draw, show, contour

from landlab import RasterModelGrid


def set_flux_coefficients(mg, dx):
    """
    Sets the spatial pattern of "K" (diffusion coefficient at cell
    faces)
    """

    # here we set it up such that K ranges from 0 at the edge of
    # cell 1 (left edge) to 1 at right edge of last cell (x=L)
    Kmax = 1.0
    Kexp = 1.0
    yf = mg.get_face_y_coords()
    xf = mg.get_face_x_coords()

    # Exponential decline from a point at center top
    decay_scale = 0.5
    xf = xf - dx/2.0
    yf = yf - dx/2.0
    x0 = 0.5 * (mg.number_of_node_columns - 2) * dx
    y0 = (mg.number_of_node_columns - 2) * dx
    dist = np.sqrt((xf - x0) ** 2.0 + (yf - y0) ** 2.0)
    K = Kmax * np.exp(-dist / decay_scale)

    return K


def main():
    # Initialize

    # User-defined parameters
    nr = 12     # number of rows
    nc = 12     # number of columns
    dx = 0.5   # cell spacing
    s0 = 0.001  # source

    # Setup
    mg = RasterModelGrid(nr, nc, dx)  # Create the grid
    u = mg.zeros(centering='cell') # Dependent variable (temperature)
    interior_cells = mg.get_interior_cells()  # ID's of interior cells
    dudt = mg.zeros(centering='cell') # Rate of change of temperature
    s = mg.zeros(centering='cell') # Source term (e.g., radioactive decay)
    s[interior_cells] = s0  # Set source at interior cells
    opt_plot = True  # Option for plotting output

    # Radially symmetric, exponential decay
    k = set_flux_coefficients(mg, dx)       # Set up diffusion coefficients

    dt = 0.25*dx**2.0/max(k)                  # Set time step
    run_time = 100.0                          # Set run time
    nt = int(round(run_time / dt))        # number of iterations

    mg.set_noflux_boundaries(True, True, False, True)  # Boundary conditions

    #-------------------------------------------------------------------
    # Process

    for i in range(0, nt):

        print i
        g = mg.calculate_face_gradients(u)  # Thermal gradients
        q = -k*g  # Heat flux across faces
        dqds = mg.calculate_flux_divergences(q)  # Divergence of heat flux
        dudt = s - dqds  # Rate of change of temperature
        u = u + dudt * dt  # Update temperature field
        u = mg.update_noflux_boundaries(u)  # Update boundaries

    #-------------------------------------------------------------------
    # Finalize

    if opt_plot:
        ur = mg.cell_vector_to_raster(u)   # Get a raster temperature field
        contour(ur)
        show()


if __name__ == '__main__':
    main()
